import logging
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple

import higra as hg
import numpy as np
from numpy.typing import ArrayLike
from skimage import measure, segmentation
from skimage.measure._regionprops import RegionProperties

from ultrack.core.segmentation.vendored.graph import mask_to_graph
from ultrack.core.segmentation.vendored.node import Node

LOG = logging.getLogger(__name__)


def _cached(f: Callable) -> Callable:
    """Hierarchy data cache"""

    @wraps(f)
    def wrapper(obj: object) -> Any:
        name = f.__name__
        if name in obj._cache:
            return obj._cache[name]
        return f(obj)

    return wrapper


class Hierarchy:
    def __init__(
        self,
        props: RegionProperties,
        hierarchy_fun: Callable = hg.watershed_hierarchy_by_area,
        cut_threshold: Optional[float] = None,
        min_area: int = 50,
        max_area: int = 1000000,
        min_frontier: float = 0.0,
        anisotropy_pen: float = 0.0,
        cache: bool = True,
    ):
        """
        Helper class to interact with the watershed hierarchy of segments.

        Parameters
        ----------
        props : RegionProperties
            Region properties computed from scikit-image `measure.regionprops`.
        hierarchy_fun : Callable, optional
            Hierarchical watershed criterion function. Defaults to hg.watershed_hierarchy_by_area.
        cut_threshold : Optional[float], optional
            Hierarchy horizontal cut threshold. Defaults to None, returning a single labeling (no cut).
        min_area : int, optional
            Minimum number of pixels per segment. Defaults to 50.
        max_area : int, optional
            Maximum number of pixels per segment. Defaults to 1000000.
        min_frontier : float, optional
            Minimum required edge weight for splitting segments. Defaults to 0.0.
        anisotropy_pen : float, optional
            Edge weight penalization for the z-axis. Defaults to 0.0.
        cache : bool, optional
            Enables cache of data structures. Defaults to True.
            Can lead to segmentation fault when False if there's tie-zones on the watershed hierarchy.
            Due to tree and altitute miss match.
        """
        self.props = props
        self.hierarchy_fun = hierarchy_fun
        self.cut_threshold = cut_threshold
        self._min_area = min_area
        self._max_area = max_area
        self._min_frontier = min_frontier
        self._anisotropy_pen = anisotropy_pen

        self._cache = {}
        self._cache_active = cache

        self._mask_buffer: Optional[ArrayLike] = None
        self._tree_buffer: Optional[ArrayLike] = None

        self._nodes = {}

    def label_buffer(self, buffer: ArrayLike, offset: int = 1) -> int:
        """
        Fills the buffer with labels given an initial offset.

        Args:
            buffer (ArrayLike): Array to be filled, must match the original image shape.
            offset (int, optional): Labeling offset. Defaults to 1.

        Returns:
            int: Number of labeled segments.
        """
        if self.cut_threshold is None:
            buffer[self.props.slice][self.props.image] = offset
            return 1

        # cache must be on for cut
        cache_status = self.cache
        self.cache = True

        cut = self.cut.horizontal_cut_from_altitude(self.cut_threshold)
        relabeled, _, _ = segmentation.relabel_sequential(
            cut.labelisation_leaves(self.tree), offset=offset
        )
        buffer[self.props.slice][self.props.image] = relabeled

        self.cache = cache_status

        return len(cut.nodes())

    @property
    def cache(self) -> bool:
        return self._cache_active

    @cache.setter
    def cache(self, status: bool) -> None:
        self._cache_active = status
        self.props._cache_active = status
        if not status:
            hg.clear_auto_cache(reference_object=self.tree)
            self._cache.clear()
            self.props._cache.clear()
            self._nodes.clear()

    @staticmethod
    def _filter_contour_strength(
        tree: hg.Tree,
        alt: ArrayLike,
        frontier: ArrayLike,
        threshold: float,
        max_area: float,
    ) -> Tuple[hg.Tree, ArrayLike]:

        LOG.info("Filtering hierarchy by contour strength.")
        irrelevant_nodes = frontier < threshold

        if max_area is not None:
            # Avoid filtering nodes where merge leads to a node with maximum area above threshold
            parent_area = hg.attribute_area(tree)[tree.parents()]
            irrelevant_nodes[parent_area > max_area] = False

        tree, node_map = hg.simplify_tree(tree, irrelevant_nodes)
        return tree, alt[node_map], frontier[node_map]

    def watershed_hierarchy(self) -> Tuple[hg.Tree, ArrayLike, ArrayLike]:
        """
        Creates and filters the watershed hierarchy.

        Raises:
            RuntimeError: It returns a error to avoid SegFault from Higra
            when processing with a single node. Increase `min_area` to avoid this.

        Returns:
            Tuple[hg.Tree, ArrayLike]:
            It returns the watershed hierarchy (tree), its altitudes.
        """
        image = self.props.intensity_image
        if image.dtype == np.float16:
            image = image.astype(np.float32)

        if image.size < 8:
            raise RuntimeError(f"Region too small. Size of {image.size} found.")

        LOG.info("Creating graph from mask.")
        mask = self.props.image
        graph, weights = mask_to_graph(mask, image, self._anisotropy_pen)

        LOG.info("Constructing hierarchy.")
        tree, alt = self.hierarchy_fun(graph, weights)

        LOG.info("Filtering small nodes of hierarchy.")
        tree, alt = hg.filter_small_nodes_from_tree(tree, alt, self._min_area)

        hg.set_attribute(graph, "no_border_vertex_out_degree", None)
        frontier = hg.attribute_contour_strength(tree, weights)
        hg.set_attribute(graph, "no_border_vertex_out_degree", 2 * mask.ndim)

        if self._min_frontier > 0.0:
            tree, alt, frontier = self._filter_contour_strength(
                tree,
                alt,
                frontier,
                self._min_frontier,
                self._max_area,
            )

        LOG.info("Filtering large nodes of hierarchy.")
        tree, node_map = hg.simplify_tree(
            tree, hg.attribute_area(tree) > self._max_area
        )
        alt = alt[node_map]
        frontier = frontier[node_map]

        return tree, alt, frontier

    @property
    @_cached
    def area(self) -> ArrayLike:
        area = hg.attribute_area(self.tree)
        if self.cache:
            self._cache["area"] = area
        return area

    @property
    @_cached
    def dynamics(self) -> ArrayLike:
        dynamics = hg.attribute_dynamics(self.tree, self.alt)
        if self.cache:
            self._cache["dynamics"] = dynamics
        return dynamics

    @property
    @_cached
    def height(self) -> ArrayLike:
        height = hg.attribute_height(self.tree, self.alt)
        if self.cache:
            self._cache["height"] = height
        return height

    @property
    @_cached
    def tree(self) -> hg.Tree:
        tree, alt, frontier = self.watershed_hierarchy()
        if self.cache:
            self._cache["tree"] = tree
            self._cache["alt"] = alt
            self._cache["frontier"] = frontier
        return tree

    @property
    @_cached
    def alt(self) -> ArrayLike:
        tree, alt, frontier = self.watershed_hierarchy()
        if self.cache:
            self._cache["tree"] = tree
            self._cache["alt"] = alt
            self._cache["frontier"] = frontier
        return alt

    @property
    @_cached
    def frontier(self) -> hg.Tree:
        tree, alt, frontier = self.watershed_hierarchy()
        if self.cache:
            self._cache["tree"] = tree
            self._cache["alt"] = alt
            self._cache["frontier"] = frontier
        return frontier

    @property
    @_cached
    def cut(self) -> hg.HorizontalCutExplorer:
        cut = hg.HorizontalCutExplorer(self.tree, self.alt)
        if self.cache:
            self._cache["cut"] = cut
        return cut

    @staticmethod
    def create_node(node_index: int, parent: "Hierarchy", **kwargs) -> Node:
        """
        Helper function for creating nodes so it can be easily overwritten when inheriting
        the Hierarchy class.

        Args:
            node_index (int): Index of the node in the original hierarchy.
            parent (Hierarchy): Hierarchy containing the node.

        Returns:
            Node: Node object or a subclass of it.
        """
        return Node(node_index, parent=parent, **kwargs)

    def compute_nodes(self) -> None:
        """
        Compute the nodes from the hierarchy given the provided
        parameters (`min_area`, `max_area`, etc.)
        """
        tree = self.tree
        area = self.area
        frontier = self.frontier
        height = self.height

        for node_idx in tree.leaves_to_root_iterator(include_leaves=False):
            if area[node_idx] > self._max_area:
                continue
            self._nodes[node_idx] = self.create_node(
                node_idx,
                self,
                area=area[node_idx].item(),
                frontier=frontier[node_idx].item(),
                height=height[node_idx].item(),
            )

    def _fix_empty_nodes(self) -> None:
        """
        If no node is found due to the filtering operations,
        their root is added as a node to avoid an empty hierarchy.
        """
        if len(self._nodes) > 0:
            return

        root_index = self.tree.root()
        self._nodes[root_index] = self.create_node(
            root_index,
            self,
            area=self.props.area,
            frontier=-1.0,
            height=-1.0,
        )

    @property
    def nodes(self) -> List[Node]:
        if len(self._nodes) == 0:
            # avoiding multiple allocation during node computation
            self._mask_buffer = np.zeros(self.props.image.shape, dtype=int)
            self._tree_buffer = np.ones(self.tree.num_vertices(), dtype=bool)

            self.compute_nodes()
            self._fix_empty_nodes()

            self._mask_buffer = None
            self._tree_buffer = None

        return list(self._nodes.values())


def oversegment_components(
    labels: ArrayLike, boundaries: ArrayLike, max_area: int, anisotropy_pen: float = 0.0
) -> ArrayLike:
    """
    This function oversegment segments given an maximum area to decrease the overall hierarchy volume (area),
    speeding up the remaining computation.

    Parameters
    ----------
    labels : ArrayLike
        Input labels to be split into smaller segments.
    boundaries : ArrayLike
        Image graph edge weights.
    max_area : int
        Maximum area (hierarchy threshold cut).
    anisotropy_pen : float, optional
        Edge weight penalization for the z-axis. Defaults to 0.0.

    Returns
    -------
    ArrayLike
        Oversgmented labels given max area paremeter.
    """
    dtype = np.promote_types(np.float32, boundaries.dtype)
    boundaries = boundaries.astype(dtype)  # required by numba
    offset = 1
    new_labels = np.zeros_like(labels)
    for c in measure.regionprops(labels, boundaries):
        if c.area > max_area:
            graph, weights = mask_to_graph(c.image, c.intensity_image, anisotropy_pen)
            tree, alt = hg.watershed_hierarchy_by_area(graph, weights)
            cut = hg.labelisation_horizontal_cut_from_threshold(tree, alt, max_area)
            cut, _, _ = segmentation.relabel_sequential(cut, offset=offset)
            offset = cut.max() + 1
            new_labels[c.slice][c.image] = cut
        else:
            new_labels[c.slice][c.image] = offset
            offset += 1
    return new_labels


def to_labels(
    hierarchies: List[Hierarchy], shape: Tuple[int], dtype: np.dtype = np.int32
) -> None:
    labels = np.zeros(shape, dtype=dtype)

    offset = 1
    for h in hierarchies:
        offset += h.label_buffer(labels, offset=offset)

    return labels



import logging
from pathlib import Path
from typing import Optional, Sequence

import click
import cloudpickle
import toml
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel
from rich.logging import RichHandler

from ultrack.cli.segment import _get_layer_data
from ultrack.cli.utils import (
    batch_index_option,
    config_option,
    napari_reader_option,
    overwrite_option,
    paths_argument,
    persistence_option,
)
from ultrack.config import MainConfig
from ultrack.core.match_gt import match_to_ground_truth
from ultrack.ml.classification import fit_links_prob, fit_nodes_prob

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(RichHandler())


@click.command("match_gt")
@paths_argument()
@napari_reader_option()
@config_option()
@click.option(
    "--ground-truth-layer",
    "-gl",
    required=False,
    type=str,
    default=None,
    help="Ground-truth layer index on napari.",
)
@click.option(
    "--output-model",
    "-om",
    type=click.Path(dir_okay=False, path_type=Path),
    required=False,
    default=None,
    help="Optional output model file path.",
)
@click.option(
    "--output-config",
    "-oc",
    type=click.Path(dir_okay=False, path_type=Path),
    help="Optional output config file path.",
)
@click.option(
    "--is-segmentation",
    is_flag=True,
    type=bool,
    default=False,
    help="Indicates ground-truth are fully curated segmentation masks. "
    "When activated different costs are used for insertions and deletions.",
)
@click.option(
    "--is-tracking",
    is_flag=True,
    type=bool,
    default=False,
    help="Indicates ground-truth are tracking instances results.",
)
@click.option(
    "--is-dense",
    is_flag=True,
    type=bool,
    default=False,
    help="Indicates ground-truth are dense annotations (everything is annotated).",
)
@click.option(
    "--insert-prob",
    is_flag=True,
    type=bool,
    default=False,
    help="Insert estimated probabilities into the database.",
)
@batch_index_option()
@overwrite_option()
@persistence_option()
def match_gt_cli(
    paths: Sequence[Path],
    reader_plugin: str,
    config: MainConfig,
    ground_truth_layer: Optional[str],
    output_model: Optional[Path],
    output_config: Optional[Path],
    is_segmentation: bool,
    is_tracking: bool,
    is_dense: bool,
    insert_prob: bool,
    batch_index: Optional[int],
    overwrite: bool,
    persistence: bool,
) -> None:
    """
    Match ground-truth labels to the segmentation/tracking database.
    """

    if output_model is not None and output_model.exists() and not overwrite:
        raise FileExistsError(
            f"Output model {output_model} already exists. Use --overwrite to overwrite."
        )

    if output_config is not None:
        if not is_segmentation:
            raise ValueError(
                "Output config is only available for segmentation ground-truth `--is-segmentation`."
            )

        if output_config.exists() and not overwrite:
            raise FileExistsError(
                f"Output config {output_config} already exists. Use --overwrite to overwrite."
            )

    # Data loading
    _initialize_plugins()

    viewer = ViewerModel()
    viewer.open(path=paths, plugin=reader_plugin)

    if ground_truth_layer is None:
        if len(viewer.layers) > 1:
            raise ValueError(
                "Multiple layers found, please specify `--ground-truth-layer`."
            )
        else:
            ground_truth_layer = viewer.layers[0].name

    gt = _get_layer_data(viewer, ground_truth_layer)

    # Match ground-truth to database
    gt_df, new_config = match_to_ground_truth(
        config=config,
        gt_labels=gt,
        scale=config.data_config.metadata.get("scale"),
        is_segmentation=is_segmentation,
        optimize_config=True,
        batch_index=batch_index,
    )

    if output_config is not None:
        LOG.info("Estimated new config: %s", new_config)
        LOG.info("Saving new config to %s", output_config)
        with open(output_config, "w") as f:
            toml.dump(new_config.model_dump(by_alias=True), f)

    if insert_prob or output_model is not None:
        model = fit_nodes_prob(
            config,
            gt_df["gt_track_id"],
            persistence_features=persistence,
            insert_prob=insert_prob,
            remove_no_overlap=not is_dense,
        )

        if is_tracking:
            link_model = fit_links_prob(
                config,
                gt_df["gt_track_id"],
                persistence_features=persistence,
                insert_prob=insert_prob,
            )
            model = {"nodes": model, "links": link_model}

        if output_model is not None:
            LOG.info("Saving model to %s", output_model)
            with open(output_model, "wb") as f:
                cloudpickle.dump(model, f)



from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from skimage.measure import regionprops_table
from toolz import curry

from ultrack.utils.array import check_array_chunk
from ultrack.utils.multiprocessing import multiprocessing_apply


def intensity_sum(mask: np.ndarray, intensities: np.ndarray) -> float:
    """Sums the intensity inside the mask"""
    return intensities[mask].sum(axis=0, dtype=float)


def intensity_std(mask: np.ndarray, intensities: np.ndarray) -> float:
    """Standard deviation of the intensity inside the mask"""
    return intensities[mask].std(axis=0, dtype=float)


def num_pixels(mask: np.ndarray) -> int:
    """Number of pixels in the mask.
    NOTE: should be removed in the future, this was a bug in scikit-image
          https://github.com/scikit-image/scikit-image/issues/7038
    """
    return mask.sum()


@curry
def _extract_measurements(
    t: int,
    segments: ArrayLike,
    image: Optional[Union[List[ArrayLike], ArrayLike]],
    channel_axis: Optional[int],
    properties: Tuple[str, ...],
    extra_properties: Optional[List],
    spatial_scale: Optional[ArrayLike],
) -> pd.DataFrame:
    """
    Extract measurements for a single time point.
    See `tracks_properties` for parameter descriptions.
    """
    if image is not None:
        # create image with channel as the last dimension
        if isinstance(image, list):
            t_image = np.stack([np.asarray(channel[t]) for channel in image], axis=-1)
        else:
            t_image = np.asarray(image[t])
            if channel_axis is not None:
                axes = list(range(t_image.ndim))
                axes.remove(channel_axis)
                axes.append(channel_axis)
                t_image = t_image.transpose(axes)
    else:
        t_image = None

    df = pd.DataFrame(
        regionprops_table(
            np.asarray(segments[t]),
            intensity_image=t_image,
            properties=properties,
            separator="_",
            extra_properties=extra_properties,
            spacing=spatial_scale,
        ),
    )

    df["t"] = t

    return df


def tracks_properties(
    segments: ArrayLike,
    tracks_df: Optional[pd.DataFrame] = None,
    image: Optional[Union[List[ArrayLike], ArrayLike]] = None,
    properties: Tuple[str, ...] = (
        "label",
        "num_pixels",
        "area",
        "intensity_sum",
        "intensity_mean",
        "intensity_std",
        "intensity_min",
        "intensity_max",
    ),
    channel_axis: Optional[int] = None,
    scale: Optional[ArrayLike] = None,
    n_workers: int = 1,
) -> pd.DataFrame:
    """
    Calculate properties of tracked regions from segmentation data.

    Parameters
    ----------
    segments : ArrayLike
        Array-like object containing the segmented regions for each time point.
        Time must be the first dimension.
    tracks_df : Optional[pd.DataFrame], default None
        DataFrame containing tracking information for regions.
        When provided, measurements are merged into this data frame.
    image : Optional[Union[List[ArrayLike], ArrayLike]], default None
        Array-like object containing the image data for each time point. If provided,
        intensity-based properties will be calculated; otherwise, only geometric
        properties will be calculated.
    properties : Tuple[str, ...], default ('num_pixels', 'area', 'intensity_sum',
                                          'intensity_mean', 'intensity_std', 'intensity_min',
                                          'intensity_max')
        Tuple of property names to be calculated. Available options include:
            - 'num_pixels': Number of pixels in the region.
            - 'area': Area of the region.
            - 'intensity_sum': Sum of pixel intensities within the region.
            - 'intensity_mean': Mean pixel intensity within the region.
            - 'intensity_std': Standard deviation of pixel intensities within the region.
            - 'intensity_min': Minimum pixel intensity within the region.
            - 'intensity_max': Maximum pixel intensity within the region.
    channel_axis : Optional[int], default None
        If the `image` parameter is provided and it contains multi-channel data, this
        parameter specifies the axis containing the channels. If None, the data is assumed
        to be single-channel.
    scale : Optional[ArrayLike], default None
        Array-like object containing the scaling factors for each dimension of segments.
        Must include time dimension.
        Used for converting pixel measurements to physical units.
    n_workers : int, default 1
        Number of workers to use for parallel processing.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the calculated properties for each region at each time point.

    Notes
    -----
    - The function uses regionprops_table from skimage.measure to calculate properties.
    - If `image` is None, only geometric properties will be calculated.
    - If `image` is provided, intensity-based properties will be calculated, provided they
      are included in the `properties` parameter.

    Examples
    --------
    >>> # Example usage with image data
    >>> image_data = ...  # Provide image data here
    >>> segments_data = ...  # Provide segmentation data here
    >>> result_df = tracks_properties(segments_data, image=image_data)

    >>> # Example usage without image data
    >>> segments_data = ...  # Provide segmentation data here
    >>> result_df = tracks_properties(segments_data)
    """
    if scale is None:
        spatial_scale = None
    else:
        # excluding non-spatial dimensions
        spatial_scale = scale[-segments.ndim + 1 :]

    rename_map = {"label": "track_id"}
    extra_properties = []

    if "num_pixels" in properties:
        extra_properties.append(num_pixels)

    if image is None:
        properties = tuple(p for p in properties if "intensity" not in p)

    else:
        if not isinstance(image, list):
            if image.ndim > segments.ndim and channel_axis is None:
                raise ValueError(
                    "Channel axis must be specified when multi-channel image is provided."
                )
            check_array_chunk(image)
        else:
            for channel in image:
                check_array_chunk(channel)

        if "intensity_sum" in properties:
            extra_properties.append(intensity_sum)

        if "intensity_std" in properties:
            extra_properties.append(intensity_std)

    extract_measurements = _extract_measurements(
        segments=segments,
        image=image,
        channel_axis=channel_axis,
        properties=properties,
        extra_properties=extra_properties,
        spatial_scale=spatial_scale,
    )

    measures = multiprocessing_apply(
        extract_measurements,
        range(segments.shape[0]),
        n_workers=n_workers,
        desc="Extracting measurements",
    )

    measures = pd.concat(measures)
    measures.rename(columns=rename_map, inplace=True)
    measures.reset_index(inplace=True, drop=True)

    if tracks_df is not None:
        measures = tracks_df.merge(
            measures,
            on=["t", "track_id"],
            how="left",
        )

    return measures



from pathlib import Path
from typing import Any, Dict

import pytest
import toml

from ultrack.config import load_config


def _assert_input_in_target(input: Dict, target: Dict) -> None:
    """Asserts input values are in target recursivelly"""
    for k in input:
        if isinstance(input[k], dict):
            _assert_input_in_target(input[k], target[k])
        else:
            assert target[k] == input[k], f"Key {k} not equal"


def _format_config(config: Dict) -> None:
    """Formats dictionary config according to Config transforms."""
    config["data_config"] = config.pop("data")
    config["segmentation_config"] = config.pop("segmentation")
    config["linking_config"] = config.pop("linking")
    config["tracking_config"] = config.pop("tracking")


def test_config_content(config_path: Path, config_content: Dict[str, Any]) -> None:
    """Tests if content is loaded correctly"""
    config = load_config(config_path)
    _format_config(config_content)
    _assert_input_in_target(config_content, config.model_dump())


def test_invalid_config_content(tmp_path: Path, config_content: Dict[str, Any]) -> None:
    """Tests invalid content"""
    config_content["segmentation"]["ws_hierarchy"] = "other"
    path = tmp_path / "config.toml"

    with open(path, mode="w") as f:
        toml.dump(config_content, f)

    with pytest.raises(ValueError):
        load_config(path)



import time
from pathlib import Path
from typing import Callable, List

import higra as hg
import napari
import pytest

from ultrack.config import MainConfig
from ultrack.widgets.ultrackwidget import UltrackWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


def test_ultrack_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    zarr_dataset_paths: List[str],
    tmp_path: Path,
) -> None:

    viewer = make_napari_viewer()
    widget = UltrackWidget(viewer)

    layers = viewer.open(zarr_dataset_paths)  # foreground, edge

    # setting combobox choices manually, because they were not working automatically
    widget._main_config_w._foreground_layer_w.choices = layers
    widget._main_config_w._edge_layer_w.choices = layers

    # selecting layers
    widget._main_config_w._foreground_layer_w.value = layers[0]
    widget._main_config_w._edge_layer_w.value = layers[1]

    # checking if widget changes are propagated to children config from multiple interfaces
    config = MainConfig()
    config.segmentation_config.threshold = 0.65

    # temporary working dir
    config.data_config.working_dir = tmp_path

    widget.config = config
    assert widget._segmentation_w._attr_to_widget["threshold"].value == 0.65
    assert widget._segmentation_w.config.threshold == 0.65
    assert widget.config.segmentation_config.threshold == 0.65

    widget._segmentation_w._attr_to_widget["threshold"].value = 0.42
    assert widget.config.segmentation_config.threshold == 0.42
    assert widget._segmentation_w.config.threshold == 0.42

    # checking if func combo is working
    widget._segmentation_w._attr_to_widget["ws_hierarchy"].value = (
        hg.watershed_hierarchy_by_dynamics
    )
    assert (
        widget.config.segmentation_config.ws_hierarchy
        == hg.watershed_hierarchy_by_dynamics
    )
    # test save config
    widget._main_config_w._config_loader_w.value = tmp_path / "config.toml"
    widget._main_config_w._save_config_btn.clicked.emit()

    delay = 4  # delay necessary due to thread worker, blocking while loop did not work. I don't know why

    widget._segmentation_w._segment_btn.clicked.emit()
    time.sleep(delay)

    widget._linking_w._link_btn.clicked.emit()
    time.sleep(delay)

    widget._tracking_w._track_btn.clicked.emit()
    time.sleep(delay)

    # checking if the whole thing can be run twice --- overwrite is working
    widget._segmentation_w._segment_btn.clicked.emit()
    time.sleep(delay)

    # checking linking with features edge weight
    widget._linking_w._images_w.value = True
    viewer.layers.selection.select_only(viewer.layers["contours"])

    widget._linking_w._link_btn.clicked.emit()
    time.sleep(delay)

    widget._tracking_w._track_btn.clicked.emit()
    time.sleep(delay)



from typing import Tuple

import pytest
import zarr

from ultrack import track
from ultrack.config import MainConfig


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 128, "n_dim": 2}),
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 64, "n_dim": 3}),
    ],
    indirect=True,
)
def test_tracking(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    foreground, contours, labels = timelapse_mock_data

    track(config_instance, labels=labels)
    track(config_instance, foreground=foreground, contours=contours, overwrite=True)



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
import scipy.ndimage as ndi
import zarr
from numpy.typing import ArrayLike
from tqdm import tqdm
from zarr.storage import StoreLike

from ultrack.utils.array import create_zarr


@dataclass
class SegmentationChange:
    src_time_pt: int
    src_label: int
    dst_time_pt: int
    dst_label: int
    shift: ArrayLike
    bbox: Optional[ArrayLike] = None
    mask: Optional[ArrayLike] = None

    def src_mask_indices(self, include_time_pt: bool = False) -> Tuple[np.ndarray, ...]:
        """
        Get the indices of the source mask in the segmentation.

        Parameters
        ----------
        include_time_pt : bool
            If True, includes the time point in the indices.

        Returns
        -------
        Tuple[np.ndarray, ...]
            The indices of the source mask.
        """
        indices = list(np.nonzero(self.mask))
        for i, b in enumerate(self.bbox[: len(indices)]):
            indices[i] = (indices[i] + b).astype(int)

        if include_time_pt:
            return (np.full_like(self.src_time_pt, len(indices[i])), *indices)
        else:
            return tuple(indices)

    def dst_mask_indices(self, include_time_pt: bool = False) -> Tuple[np.ndarray, ...]:
        """
        Get the indices of the destination mask in the segmentation.

        Parameters
        ----------
        include_time_pt : bool
            If True, includes the time point in the indices.

        Returns
        -------
        Tuple[np.ndarray, ...]
            The indices of the destination mask.
        """
        indices = list(self.src_mask_indices())
        for i, s in enumerate(self.shift):
            indices[i] = (indices[i] + s).astype(int)

        if include_time_pt:
            return (np.full_like(indices[i], self.dst_time_pt), *indices)
        else:
            return tuple(indices)

    def __str__(self) -> str:
        return (
            "Segmentation update\n"
            f"from src_time_pt={self.src_time_pt} to dst_time_pt={self.dst_time_pt}\n"
            f"from src_label={self.src_label} to dst_label={self.dst_label}\n"
            f"with shift {self.shift}"
        )

    def __repr__(self) -> str:
        return str(self)


class SegmentationPainter:
    """
    Class to apply in-place changes to a segmentation time lapse.

    Parameters
    ----------
    segments : ArrayLike
        The segmentation to be updated.
    """

    def __init__(self, segments: ArrayLike) -> None:
        self._segments = segments
        self.src_tp_to_changes: Dict[int, List[SegmentationChange]] = defaultdict(list)
        self.dst_tp_to_changes: Dict[int, List[SegmentationChange]] = defaultdict(list)
        self._remap: Dict[Tuple[int, int], int] = {}

    def add_relabel(self, time_pt: int, src_label: int, dst_label: int) -> None:
        """
        Add a change in the segmentation at the given time point to the queue.

        Parameters
        ----------
        time_pt : int
            The time point at which the change occurred.
        src_label : int
            The previous label of the segment.
        dst_label : int
            The new label of the segment.
        """
        time_pt = int(time_pt)
        src_label = int(src_label)
        dst_label = int(dst_label)

        # check if it's accumulating changes
        for change in self.dst_tp_to_changes[time_pt]:
            # accumulate changes and returns without adding
            # this could happen in consecutive changes to the same label
            if change.dst_label == src_label:
                change.dst_label = dst_label
                return

        segm_change = SegmentationChange(
            time_pt,
            src_label,
            time_pt,
            dst_label,
            np.zeros(self._segments.ndim - 1),
        )

        self.src_tp_to_changes[time_pt].append(segm_change)
        self.dst_tp_to_changes[time_pt].append(segm_change)

    def add_new(
        self,
        src_time_pt: int,
        src_label: int,
        dst_time_pt: int,
        dst_label: int,
        shift: ArrayLike,
    ) -> None:
        """
        Add a new segment to the queue.

        Parameters
        ----------
        src_time_pt : int
            The time point of the reference segment.
        src_label : int
            The label of the reference segment.
        dst_time_pt : int
            The time point of the new segment.
        dst_label : int
            The label of the new segment.
        shift : ArrayLike
            The shift of the new segment.
        """
        segm_change = SegmentationChange(
            int(src_time_pt),
            int(src_label),
            int(dst_time_pt),
            int(dst_label),
            shift,
        )

        self.src_tp_to_changes[src_time_pt].append(segm_change)
        self.dst_tp_to_changes[dst_time_pt].append(segm_change)

    def apply_changes(self) -> None:
        """
        Apply the changes to the segmentation.

        It's optimized to avoid reloading the same time point multiple times.
        """

        # Loading source time points masks
        for t, changes in tqdm(
            self.src_tp_to_changes.items(), "Loading references masks"
        ):
            frame = np.asarray(self._segments[t])
            label_to_slicing = {
                idx + 1: slicing
                for idx, slicing in enumerate(ndi.find_objects(frame))
                if slicing is not None
            }

            # first add to dict to avoid recomputing the mask
            label_to_bbox_mask = {}
            for change in changes:
                if change.src_label in label_to_slicing:
                    # creating cache of bbox and mask for the source label
                    if change.src_label not in label_to_bbox_mask:
                        slicing = label_to_slicing[change.src_label]
                        label_to_bbox_mask[change.src_label] = (
                            slicing,
                            frame[slicing] == change.src_label,
                        )
                    slicing, change.mask = label_to_bbox_mask[change.src_label]
                    change.bbox = np.asarray(
                        [s.start for s in slicing] + [s.stop for s in slicing]
                    )
                else:
                    if frame.sum() == 0:
                        raise ValueError(f"Frame {t} is empty. Something went wrong.")

                    print(f"{change} not found in frame {t}")

        # Applying changes to destination
        for t, changes in tqdm(self.dst_tp_to_changes.items(), "Applying changes"):
            frame = np.asarray(self._segments[t])
            for change in changes:
                if change.mask is None:
                    print(f"Mask not found for {change}")
                    continue
                dst_indices = change.dst_mask_indices()
                dst_indices = tuple(
                    np.maximum(0, np.minimum(i, s - 1))
                    for i, s in zip(dst_indices, frame.shape)
                )  # clipping mask to frame
                frame[dst_indices] = change.dst_label
            self._segments[t] = frame

        self.src_tp_to_changes.clear()
        self.dst_tp_to_changes.clear()


def copy_segments(
    segments: ArrayLike,
    segments_store_or_path: StoreLike = None,
    overwrite: bool = False,
) -> zarr.Array:
    """
    Copy the segments to a new zarr array.

    Parameters
    ----------
    segments : ArrayLike
        The segments to copy.
    segments_store_or_path : Union[Store, Path, str, None]
        The store or path to save the new segments.
    overwrite : bool
        If True, overwrite the existing array.

    Returns
    -------
    zarr.Array
        The new segments array.
    """
    is_zarr = isinstance(segments, zarr.Array)
    out_segments = create_zarr(
        segments.shape,
        segments.dtype,
        segments_store_or_path,
        chunks=segments.chunks if is_zarr else None,
        overwrite=overwrite,
    )

    # FIXME: waiting for zarr v3 to implement `zarr.copy_store`
    # if is_zarr:
    #     print("Copying segments...")
    #     # not very clean because we just created the array above
    #     zarr.copy_store(segments.store, out_segments.store, if_exists="replace")
    #     out_segments = zarr.open(
    #         out_segments.store
    #     )  # not sure why this is necessary in large datasets
    # else:
    for t in tqdm(range(segments.shape[0]), "Copying segments"):
        out_segments[t] = np.asarray(segments[t])

    return out_segments



import logging

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from skimage.measure import regionprops_table
from tqdm import tqdm

from ultrack.tracks.stats import tracks_df_movement

LOG = logging.getLogger(__name__)

_RENAME_COLUMNS = {
    "centroid-0": "z",
    "centroid-1": "y",
    "centroid-2": "x",
    "label": "track_id",
}


def _to_dataframe(labels: ArrayLike) -> pd.DataFrame:
    labels = np.asarray(labels)
    if labels.ndim == 2:
        labels = np.expand_dims(labels, 0)
    return pd.DataFrame(
        regionprops_table(labels, properties=["label", "centroid", "area"])
    )


def estimate_parameters_from_labels(
    labels: ArrayLike,
    is_timelapse: bool,
) -> pd.DataFrame:
    """
    Estimates `area` (per segment) and `distance`
    (between subsequent centroids of the same label) parameters from labels.

    Parameters
    ----------
    labels : ArrayLike
        Array of labels, segments belonging to the same track must have the same label.
    is_time_lapse : bool
        Indicates if it's a timelapse or a single stack.

    Returns
    -------
    pd.DataFrame
        Dataframe with `area` and `distance` colums.
    """
    if not is_timelapse:
        df = _to_dataframe(labels)
    else:
        df = []
        for t in tqdm(range(labels.shape[0]), "Estimating params."):
            _df = _to_dataframe(labels[t])
            _df["t"] = t
            df.append(_df)

        df = pd.concat(df)

    df.columns = [_RENAME_COLUMNS.get(c, c) for c in df.columns]

    if "t" in df.columns:
        df["distance"] = np.linalg.norm(tracks_df_movement(df, lag=1), axis=1)

    return df



import logging
from typing import Callable, Optional

import numpy as np
import zarr
from numpy.typing import ArrayLike
from tqdm import tqdm
from zarr.storage import StoreLike

from ultrack.utils.array import create_zarr
from ultrack.utils.cuda import import_module, to_cpu

LOG = logging.getLogger(__name__)


def register_timelapse(
    timelapse: ArrayLike,
    *,
    store_or_path: Optional[StoreLike] = None,
    overwrite: bool = False,
    to_device: Callable[[ArrayLike], ArrayLike] = lambda x: x,
    reference_channel: Optional[int] = None,
    overlap_ratio: float = 0.25,
    normalization: Optional[str] = None,
    padding: Optional[int] = None,
    **kwargs,
) -> zarr.Array:
    """
    Register a timelapse sequence using phase cross correlation.

    Parameters
    ----------
    timelapse : ArrayLike
        Input timelapse sequence, T(CZ)YX array C and Z are optional.
        NOTE: when provided, C must be the second dimension after T.
    store_or_path : StoreLike, optional
        Zarr storage or a file path, to save the output, useful for larger than memory datasets.
        By default it loads the data into memory.
    overwrite : bool, optional
        Overwrite output file if it already exists, when using directory store or a path.
    to_device : Callable[[ArrayLike], ArrayLike], optional
        Function to move the input data to cuda device, by default lambda x: x (CPU).
    reference_channel : Optional[int], optional
        Reference channel for registration, by default None.
        It must be provided when it contains a channel dimension.
    overlap_ratio : float, optional
        Overlap ratio for phase cross correlation, by default 0.25.
    normalization : Optional[str], optional
        Normalization method for phase cross correlation, by default None.
    padding : Optional[int], optional
        Padding for registration, by default None.
    **kwargs
        Additional arguments for phase cross correlation. See `skimage.registration phase_cross_correlation
        <https://scikit-image.org/docs/stable/api/skimage.registration.html#skimage.registration.phase_cross_correlation>`_.  # noqa: E501

    Returns
    -------
    zarr.Array
        Registered timelapse sequence.
    """
    shape = list(timelapse.shape)

    if padding is not None:
        offset = 1 if reference_channel is None else 2
        pads = [(0, 0)] * (offset - 1)

        for i in range(offset, len(shape)):
            shape[i] += 2 * padding
            pads.append((padding, padding))

        def maybe_pad(x: np.ndarray) -> np.ndarray:
            x = np.asarray(x)
            x = to_device(x)
            return np.pad(x, pads, mode="constant")

    else:

        def maybe_pad(x: np.ndarray) -> np.ndarray:
            x = np.asarray(x)
            return to_device(x)

    out_arr = create_zarr(
        tuple(shape),
        dtype=timelapse.dtype,
        store_or_path=store_or_path,
        overwrite=overwrite,
    )

    if reference_channel is None:
        channel = ...
    else:
        channel = reference_channel

    prev_frame = maybe_pad(timelapse[0])
    out_arr[0] = to_cpu(prev_frame)

    ndi = import_module("scipy", "ndimage", arr=prev_frame)
    skreg = import_module("skimage", "registration", arr=prev_frame)

    for t in tqdm(range(timelapse.shape[0] - 1), "Registration"):
        next_frame = maybe_pad(timelapse[t + 1])
        shift, _, _ = skreg.phase_cross_correlation(
            prev_frame[channel],
            next_frame[channel],
            overlap_ratio=overlap_ratio,
            normalization=normalization,
            **kwargs,
        )

        LOG.info("Shift at {t}: {shift}", t=t, shift=shift)
        print(f"Shift at {t}: {shift}")

        if reference_channel is not None:
            shift = (0, *shift)

        next_frame = ndi.shift(next_frame, shift, order=1)
        out_arr[t + 1] = to_cpu(next_frame)

        prev_frame = next_frame

    return out_arr



from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
import pytest
from napari.viewer import ViewerModel

from ultrack.tracks.video import (
    tracks_df_to_3D_video,
    tracks_df_to_moving_2D_plane_video,
    tracks_df_to_videos,
)
from ultrack.utils.constants import NO_PARENT

pytest.importorskip("napari_animation")


@pytest.fixture
def tracks_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "parent_id": [NO_PARENT, 0, 1, 2, 1, 4],
            "track_id": [1, 1, 2, 2, 3, 3],
            "parent_track_id": [NO_PARENT, NO_PARENT, 1, 1, 1, 1],
            "t": [0, 1, 2, 3, 2, 3],
            "z": [0, 0, 0, 0, 0, 0],
            "y": [10, 20, 30, 10, 20, 10],
            "x": [10, 20, 30, 10, 20, 10],
        }
    )


def test_tracks_video(
    tracks_df: pd.DataFrame,
    make_napari_viewer: Callable[[], ViewerModel],
    tmp_path: Path,
) -> None:

    image = np.random.randint(255, size=(5, 64, 64))

    viewer = make_napari_viewer()
    viewer.add_image(image)

    tracks_df_to_videos(
        viewer,
        tracks_df,
        tmp_path,
    )


def test_tracks_3d_video(
    tracks_df: pd.DataFrame,
    make_napari_viewer: Callable[[], ViewerModel],
    tmp_path: Path,
) -> None:

    image = np.random.randint(255, size=(5, 64, 64, 64))

    viewer = make_napari_viewer()
    im_layer = viewer.add_image(image)

    tracks_df_to_3D_video(
        viewer,
        tracks_df,
        tmp_path / "video.mp4",
        keyframe_step=1,
        update_step=1,
        clipping_planes_layers=[im_layer],
        time_to_camera_kwargs={
            0: {"zoom": 1},
            3: {"zoom": 2, "angles": "auto", "center": "auto"},
        },
    )


def test_tracks_moving_slice_video(
    tracks_df: pd.DataFrame,
    make_napari_viewer: Callable[[], ViewerModel],
    tmp_path: Path,
) -> None:
    image = np.random.randint(255, size=(5, 64, 64))

    viewer = make_napari_viewer()
    viewer.add_image(image)

    tracks_df_to_moving_2D_plane_video(
        viewer,
        tracks_df,
        tmp_path / "2D_slice_video.mp4",
    )



from pathlib import Path
from typing import Optional, Sequence

import click
import dask.array as da
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel
from numpy.typing import ArrayLike

from ultrack import segment
from ultrack.cli.utils import (
    batch_index_option,
    config_option,
    napari_reader_option,
    overwrite_option,
    paths_argument,
)
from ultrack.config import MainConfig


def _get_layer_data(viewer: ViewerModel, key: str) -> ArrayLike:
    """Get layer data from napari viewer."""
    layer = viewer.layers[key]
    if layer.multiscale:
        return layer.data[0]
    else:
        return layer.data


@click.command("segment")
@paths_argument()
@napari_reader_option()
@config_option()
@click.option(
    "--foreground-layer",
    "-fl",
    required=True,
    type=str,
    help="Cell foreground layer index on napari.",
)
@click.option(
    "--contours-layer",
    "-cl",
    required=True,
    type=str,
    help="Cell contours layer index on napari.",
)
@click.option(
    "--images-layer",
    "-il",
    required=False,
    multiple=True,
    type=str,
    help="Image layer index on napari for intensity features, it can used multiple times for multiple channels.",
)
@click.option(
    "--insertion-throttle-rate",
    required=False,
    type=int,
    help="Rate at which to insert new hierarchies (group of competing segments) into the database.",
    default=50,
)
@click.option(
    "--properties",
    "-p",
    type=str,
    multiple=True,
    help="Compute properties of the segments, it can be used multiple times for multiple properties.",
)
@batch_index_option()
@overwrite_option()
def segmentation_cli(
    paths: Sequence[Path],
    reader_plugin: str,
    config: MainConfig,
    foreground_layer: str,
    contours_layer: str,
    images_layer: Sequence[str],
    insertion_throttle_rate: int,
    properties: Sequence[str],
    batch_index: Optional[int],
    overwrite: bool,
) -> None:
    """Compute candidate segments for tracking model from input data."""
    _initialize_plugins()

    viewer = ViewerModel()
    viewer.open(path=paths, plugin=reader_plugin)

    foreground = _get_layer_data(viewer, foreground_layer)
    edge = _get_layer_data(viewer, contours_layer)

    if len(images_layer) == 0:
        images = None
    else:
        has_intensity = any("intensity" in prop for prop in properties)
        if not has_intensity:
            raise ValueError(
                "Intensity features are required to compute image intensity properties.\n"
                "Found properties: {}\n".format(properties),
                "Expected properties: intensity_mean, intensity_std, intensity_sum or intensity_min, intensity_max",
            )

        if len(images_layer) == 1:
            images = _get_layer_data(viewer, images_layer[0])
        else:
            images = da.stack(
                [_get_layer_data(viewer, key) for key in images_layer], axis=-1
            )

    if batch_index is None or batch_index == 0:
        # this is not saved inside the `segment` function because this info
        # isn't available there
        config.data_config.metadata_add(
            {"scale": viewer.layers[contours_layer].scale.tolist()}
        )

    del viewer

    segment(
        foreground,
        edge,
        config,
        batch_index=batch_index,
        overwrite=overwrite,
        insertion_throttle_rate=insertion_throttle_rate,
        images=images,
        properties=None if len(properties) == 0 else properties,
    )



from typing import Tuple

import numpy as np
import pytest
import scipy.ndimage as ndi
import zarr

from ultrack.imgproc import register_timelapse


@pytest.mark.parametrize(
    "timelapse_mock_data",
    [
        {"length": 3, "size": 32, "n_dim": 3},
    ],
    indirect=True,
)
def test_register_timelapse(
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    _, moved_edges, _ = timelapse_mock_data

    shift = 8

    # adding a new to emulate a channel
    moved_edges = moved_edges[...][:, None]

    for i in range(moved_edges.shape[0]):
        moved_edges[i] = ndi.shift(moved_edges[i], (0, i * shift // 2, 0, 0), order=1)

    fixed_edges = register_timelapse(moved_edges, reference_channel=0, padding=shift)

    for i in range(moved_edges.shape[0] - 1):
        # removing padding and out of fov regions
        volume = fixed_edges[i, :, : -2 * shift]
        next_vol = fixed_edges[i + 1, :, : -2 * shift]

        np.testing.assert_allclose(volume, next_vol)



from pathlib import Path
from typing import Callable, Tuple

import napari
import numpy as np
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import NodeDB, VarAnnotation
from ultrack.widgets import DivisionAnnotationWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "segmentation.min_area": 25,
            "tracking.division_weight": 0,
        }
    ],
    indirect=True,
)
def test_division_annotation_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    cell_division_mock_data: Tuple[np.ndarray, np.ndarray, np.ndarray],
    tracked_cell_division_mock_data: MainConfig,
    tmp_path: Path,
    request,
) -> None:
    # NOTE: Use "--show-napari-viewer" to show viewer, useful when debugging

    config = tracked_cell_division_mock_data
    config.data_config.working_dir = tmp_path

    viewer = make_napari_viewer()

    for arr in cell_division_mock_data:
        viewer.add_image(arr)

    widget = DivisionAnnotationWidget(viewer)

    viewer.window.add_dock_widget(widget)

    widget.config = config.data_config
    assert len(widget._nodes) > 0

    if request.config.getoption("--show-napari-viewer"):
        napari.run()
        return

    assert widget.list_index == 0

    widget._annot_w.value = VarAnnotation.REAL
    widget._confirm_btn.clicked.emit()
    assert widget.list_index == 1

    widget._annot_w.value = VarAnnotation.FAKE
    widget._confirm_btn.clicked.emit()
    assert widget.list_index == 1  # there are only two divisions

    engine = create_engine(widget.config.database_path)
    with Session(engine) as session:
        div_annot = [
            annot
            for annot, in session.query(NodeDB.division).where(
                NodeDB.division != VarAnnotation.UNKNOWN
            )
        ]

    assert len(div_annot) == 2
    assert VarAnnotation.REAL in div_annot
    assert VarAnnotation.FAKE in div_annot



from typing import Tuple

import numpy as np
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ultrack import MainConfig, add_new_node
from ultrack.core.database import LinkDB, NodeDB, OverlapDB


def _get_table_sizes(session: Session) -> Tuple[int, int, int]:
    return (
        session.query(NodeDB).count(),
        session.query(LinkDB).count(),
        session.query(OverlapDB).count(),
    )


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "data.database": "sqlite",
            "segmentation.n_workers": 4,
            "linking.n_workers": 4,
            "linking.max_distance": 500,  # too big and ignored
        },
    ],
    indirect=True,
)
def test_clear_solution(
    linked_database_mock_data: MainConfig,
) -> None:

    mask = np.ones((7, 12, 12), dtype=bool)
    bbox = np.array([25, 4, 42, 32, 16, 54], dtype=int)

    engine = create_engine(linked_database_mock_data.data_config.database_path)
    with Session(engine) as session:
        n_nodes, n_links, n_overlaps = _get_table_sizes(session)

        add_new_node(
            linked_database_mock_data,
            0,
            mask,
            bbox,
        )

        new_n_nodes, new_n_links, new_n_overlaps = _get_table_sizes(session)

    assert new_n_nodes == n_nodes + 1
    assert new_n_overlaps > n_overlaps
    # could smaller than max neighbors because of radius
    assert (
        new_n_links == n_links + linked_database_mock_data.linking_config.max_neighbors
    )



from pathlib import Path
from typing import Tuple
from unittest.mock import patch

import networkx as nx
import numpy as np
import pytest
import torch as th
import torch.nn.functional as F
import zarr

import ultrack
from ultrack import MainConfig, link, segment, solve, to_ctc, to_tracks_layer, track
from ultrack.core.export import tracks_layer_to_networkx, tracks_layer_to_trackmate
from ultrack.core.tracker import TrackerStatus
from ultrack.imgproc.flow import add_flow, identity_grid


@pytest.fixture
def mock_flow_field(request) -> np.ndarray:
    ndim = request.param["ndim"]
    size = request.param["size"]
    length = request.param["length"]
    theta = th.tensor(0.5)
    shape = tuple(size for _ in range(ndim))
    if ndim == 2:
        T = th.tensor(
            [[th.cos(theta), -th.sin(theta), 0], [th.sin(theta), th.cos(theta), 0]]
        )
    else:
        T = th.tensor(
            [
                [th.cos(theta), -th.sin(theta), 0, 0],
                [th.sin(theta), th.cos(theta), 0, 0],
                [0, 0, 1, 0],
            ]
        )
    T = T[None, ...]
    grid_shape = (1, 1) + shape
    mock_grid = F.affine_grid(T, grid_shape, align_corners=True)

    flow = identity_grid(shape) - mock_grid
    flow = flow.movedim(-1, 1)
    flow = th.cat([flow for _ in range(length)], dim=0)
    return flow.numpy() / 2


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 128, "n_dim": 2}),
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 64, "n_dim": 3}),
    ],
    indirect=True,
)
def test_track(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    foreground, contours, labels = timelapse_mock_data

    # 1st track call using labels
    track(config_instance, labels=labels)
    df_original, graph_original = to_tracks_layer(config_instance)

    tracker = ultrack.Tracker(config_instance)
    tracker.track(labels=labels, overwrite=True)
    df_tracker, graph_tracker = to_tracks_layer(config_instance)

    assert df_original.equals(df_tracker)
    assert graph_original == graph_tracker

    # 2nd track call using foreground and contours
    track(config_instance, foreground=foreground, contours=contours, overwrite=True)
    df_original, graph_original = to_tracks_layer(config_instance)

    tracker = ultrack.Tracker(config_instance)
    tracker.track(foreground=foreground, contours=contours, overwrite=True)
    df_tracker, graph_tracker = to_tracks_layer(config_instance)

    assert df_original.equals(df_tracker)
    assert graph_original == graph_tracker


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 128, "n_dim": 2}),
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 64, "n_dim": 3}),
    ],
    indirect=True,
)
def test_link_segment_and_solve(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    foreground, contours, _ = timelapse_mock_data

    tracker = ultrack.Tracker(config_instance)
    tracker.segment(foreground=foreground, contours=contours)
    tracker.link()
    tracker.solve()

    df, graph = to_tracks_layer(config_instance)

    segment(
        foreground=foreground, contours=contours, config=config_instance, overwrite=True
    )
    link(config_instance, overwrite=True)
    solve(config_instance, overwrite=True)

    df_original, graph_original = to_tracks_layer(config_instance)

    assert df.equals(df_original)
    assert graph == graph_original


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 128, "n_dim": 2}),
        ({"segmentation.n_workers": 4}, {"length": 4, "size": 64, "n_dim": 3}),
    ],
    indirect=True,
)
def test_outputs(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
    tmp_path: str,
) -> None:
    tmp_path = Path(tmp_path)
    foreground, contours, _ = timelapse_mock_data

    tracker = ultrack.Tracker(config_instance)
    tracker.segment(foreground=foreground, contours=contours)
    tracker.link()
    tracker.solve()

    # test to_tracks_layer (and indirectly to_pandas)
    df_tracker, graph_tracker = tracker.to_tracks_layer()
    df_original, graph_original = to_tracks_layer(config_instance)

    assert df_tracker.equals(df_original)
    assert graph_tracker == graph_original

    # test to_ctc
    tracker.to_ctc(output_dir=tmp_path)
    assert (tmp_path / "res_track.txt").exists()
    with open(tmp_path / "res_track.txt") as f:
        content = f.read()

    to_ctc(output_dir=tmp_path, config=config_instance, overwrite=True)
    assert (tmp_path / "res_track.txt").exists()
    with open(tmp_path / "res_track.txt") as f:
        content_original = f.read()

    assert content == content_original

    # test to_zarr
    z_tracker = tracker.to_zarr()
    z_original = ultrack.tracks_to_zarr(config_instance, tracks_df=df_original)

    assert np.array_equal(np.asarray(z_tracker), np.asarray(z_original))

    # test to_trackmate
    out_trackmate_original = tracks_layer_to_trackmate(df_original)
    out_trackmate_tracker = tracker.to_trackmate()

    assert out_trackmate_original == out_trackmate_tracker

    # test to_networkx
    nx_tracker = tracker.to_networkx()
    nx_original = tracks_layer_to_networkx(df_original)

    assert nx.utils.graphs_equal(nx_tracker, nx_original)

    # test to_geff
    with patch("ultrack.core.export.geff.write_arrays") as mock_write_nx:
        tracker.to_geff(tmp_path / "test.geff")
        mock_write_nx.assert_called_once()

        # Test with overwrite=True
        tracker.to_geff(tmp_path / "test.geff", overwrite=True)
        assert mock_write_nx.call_count == 2


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data,mock_flow_field",
    [
        (
            {"segmentation.n_workers": 4},
            {"length": 4, "size": 128, "n_dim": 2},
            {"size": 32, "ndim": 2, "length": 4},
        ),
        (
            {"segmentation.n_workers": 4},
            {"length": 4, "size": 64, "n_dim": 3},
            {"size": 16, "ndim": 3, "length": 4},
        ),
    ],
    indirect=True,
)
def test_flow(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
    mock_flow_field: np.ndarray,
) -> None:
    foreground, contours, _ = timelapse_mock_data

    tracker = ultrack.Tracker(config_instance)
    tracker.segment(foreground=foreground, contours=contours)
    tracker.add_flow(mock_flow_field)
    tracker.link()
    tracker.solve()

    df, graph = to_tracks_layer(config_instance)

    segment(
        foreground=foreground, contours=contours, config=config_instance, overwrite=True
    )
    add_flow(config_instance, mock_flow_field)
    link(config_instance, overwrite=True)
    solve(config_instance, overwrite=True)

    df_original, graph_original = to_tracks_layer(config_instance)

    assert df.equals(df_original)
    assert graph == graph_original


def test_error(
    config_instance: MainConfig,
) -> None:
    tracker = ultrack.Tracker(config_instance)

    try:
        tracker.link()
    except ValueError:
        pass

    assert tracker.status == TrackerStatus.NOT_COMPUTED



from pathlib import Path

import click
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import sqlalchemy as sqla
from sqlalchemy import func
from sqlalchemy.orm import Session

from ultrack.cli.utils import config_option
from ultrack.config import MainConfig
from ultrack.core.database import LinkDB, NodeDB
from ultrack.core.export.utils import solution_dataframe_from_sql
from ultrack.tracks.graph import add_track_ids_to_tracks_df
from ultrack.utils.constants import NO_PARENT
from ultrack.utils.printing import pretty_print_df


def _nodes_count_over_time(database_path: str, fig_path: Path) -> None:
    """Queries and plots nodes counts over time. Solution isn't taken into account."""
    engine = sqla.create_engine(database_path)
    with Session(engine) as session:
        query = session.query(NodeDB.t, func.count(NodeDB.t)).group_by(NodeDB.t)
        df = pd.read_sql(query.statement, session.bind, index_col="t")

    sns.set_theme(style="whitegrid")
    plot = sns.lineplot(data=df, legend=False)
    plot.set_ylabel("count")

    plot.get_figure().savefig(fig_path)
    plt.close()

    print(f"Nodes count over time saved at {fig_path}")


def q1(arr: pd.Series) -> float:
    return arr.quantile(0.25)


def q2(arr: pd.Series) -> float:
    return arr.median()


def q3(arr: pd.Series) -> float:
    return arr.quantile(0.75)


def _link_stats_over_time(database_path: str, out_dir: Path) -> None:
    """Queries and plots link statistics over time. Solution isn't taken into account."""
    engine = sqla.create_engine(database_path)
    with Session(engine) as session:
        query = session.query(NodeDB.t, LinkDB.weight).join(
            NodeDB, NodeDB.id == LinkDB.source_id
        )
        groups = pd.read_sql(query.statement, session.bind).groupby("t")

    sns.set_theme(style="whitegrid")

    fig_path = out_dir / "link_weight_plot.png"
    df = pd.melt(
        groups.agg({"weight": [q1, q2, q3]}),
        var_name="quantile",
        value_name="link_weight",
    )
    plot = sns.lineplot(data=df, legend=False)
    plot.set_ylabel("link weight")
    plot.get_figure().savefig(fig_path)
    plt.close()

    print(f"Links weights over time saved at {fig_path}")

    fig_path = out_dir / "link_count_plot.png"
    plot = sns.lineplot(data=groups.count(), legend=False)
    plot.set_ylabel("count")
    plot.get_figure().savefig(fig_path)
    plt.close()

    print(f"Links count over time saved at {fig_path}")


def _solution_summary(database_path: str) -> None:
    """Computes some statistics from the solution."""
    df = solution_dataframe_from_sql(database_path)
    df = add_track_ids_to_tracks_df(df)

    total = len(df)
    no_parents = (df["parent_id"] == NO_PARENT).sum()
    divisions = (df.groupby("parent_id").size() > 1).sum() - int(
        no_parents > 0
    )  # subtracting NO_PARENTS
    ends = total - df.index.isin(df["parent_id"]).sum()
    tracks = len(df["track_id"].unique())

    print("\n")
    pretty_print_df(
        pd.DataFrame(
            [
                [total],
                [no_parents],
                [ends],
                [divisions],
                [tracks],
            ],
            columns=["count"],
            index=["segments", "appearance", "disappearance", "division", "tracks"],
        ),
        title="Solution summary",
    )


@click.command("data_summary")
@config_option()
@click.option(
    "--output-directory",
    "-o",
    required=True,
    type=click.Path(path_type=Path),
    default="summary",
    show_default=True,
    help="Output plots directory.",
)
def data_summary_cli(
    config: MainConfig,
    output_directory: Path,
) -> None:
    """Prints a summary of the database data."""

    database_path = config.data_config.database_path
    output_directory.mkdir(exist_ok=True)

    _nodes_count_over_time(database_path, output_directory / "nodes_count.png")
    _link_stats_over_time(database_path, output_directory)
    _solution_summary(database_path)



import napari
from magicgui.widgets import Container, FloatSpinBox, PushButton, SpinBox
from napari.layers import Labels

from ultrack.utils.edge import labels_to_contours


class LabelsToContoursWidget(Container):
    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__()
        self._viewer = viewer

        self._sigma_w = FloatSpinBox(min=0, max=25, value=0, label="Smoothing")
        self.append(self._sigma_w)

        self._selection_w = SpinBox(label="Num. selected layers", enabled=False)
        self.append(self._selection_w)
        self._update_count()

        self._viewer.layers.selection.events.changed.connect(self._update_count)

        self._run_btn = PushButton(text="Convert")
        self._run_btn.clicked.connect(self._on_run)
        self.append(self._run_btn)

    def _update_count(self) -> None:
        self._selection_w.value = len(self._viewer.layers.selection)

    def _on_run(self) -> None:
        for layer in self._viewer.layers.selection:
            if not isinstance(layer, Labels):
                raise ValueError(f"Layer named `{layer.name}` isn't of `Labels` type.")

        selection = list(layer.data for layer in self._viewer.layers.selection)
        if len(selection) == 0:
            raise ValueError("Select labels for conversion.")

        sigma = self._sigma_w.value if self._sigma_w.value > 0 else None

        foreground, contours = labels_to_contours(selection, sigma)

        self._viewer.add_image(foreground, name="foreground")
        self._viewer.add_image(contours, name="contours")



from magicgui.widgets import CheckBox, FloatSpinBox, PushButton, SpinBox

from ultrack.config.config import LinkingConfig
from ultrack.widgets.ultrackwidget._legacy.baseconfigwidget import BaseConfigWidget


class LinkingWidget(BaseConfigWidget):
    def __init__(self, config: LinkingConfig):
        super().__init__(label="Linking", config=config)
        self._images_w = CheckBox(
            label="Use selected layers",
            value=False,
            tooltip="When checked, the selected layers are used to compute edge weights with "
            "discrete cosine transform, otherwise the IoU of the segments are used.",
        )
        self.append(self._images_w)

    def _setup_widgets(self) -> None:
        self._attr_to_widget = {
            "max_neighbors": SpinBox(label="Max. neighbors"),
            "max_distance": FloatSpinBox(label="Max. distance"),
            "n_workers": SpinBox(label="Num. workers"),
        }
        for widget in self._attr_to_widget.values():
            self.append(widget)

        self._link_btn = PushButton(text="Link")
        self.append(self._link_btn)



from pathlib import Path
from typing import Optional, Sequence

import click
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel

from ultrack.cli.utils import (
    napari_reader_option,
    output_directory_option,
    overwrite_option,
    paths_argument,
)
from ultrack.utils.data import validate_and_overwrite_path
from ultrack.utils.edge import labels_to_contours


@click.command("labels_to_contours")
@paths_argument()
@output_directory_option(help="`detection.zarr` and `contours.zarr` output directory.")
@napari_reader_option()
@click.option(
    "--sigma",
    "-s",
    type=float,
    default=None,
    show_default=True,
    help="Contour smoothing parameter (gaussian blur sigma). No blurring by default.",
)
@overwrite_option()
def labels_to_contours_cli(
    paths: Sequence[Path],
    output_directory: Path,
    reader_plugin: str,
    sigma: Optional[float],
    overwrite: bool,
) -> None:
    """
    Converts and merges a sequence of labels into ultrack input format (foreground and contours)
    """
    foreground_path = output_directory / "foreground.zarr"
    validate_and_overwrite_path(foreground_path, overwrite, "cli")

    contours_path = output_directory / "contours.zarr"
    validate_and_overwrite_path(contours_path, overwrite, "cli")

    _initialize_plugins()

    viewer = ViewerModel()
    viewer.open(path=paths, plugin=reader_plugin)

    labels = [layer.data for layer in viewer.layers]
    del viewer

    labels_to_contours(
        labels,
        sigma=sigma,
        foreground_store_or_path=foreground_path,
        contours_store_or_path=contours_path,
    )



import functools
import importlib
import logging
from contextlib import contextmanager
from types import ModuleType
from typing import Callable, Generator, Optional

import dask.array as da
import numpy as np
from numpy.typing import ArrayLike

LOG = logging.getLogger(__name__)


malloc_managed = None

try:
    import cupy as cp

    if not cp.cuda.is_available():
        cp = None
        LOG.info("cupy found but cuda is not available.")
    else:
        from cupy.cuda.memory import malloc_managed

        LOG.info("cupy found.")
        xp = cp

except (ModuleNotFoundError, ImportError):
    LOG.info("cupy not found.")
    cp = None
    xp = np

try:
    import torch as th

    LOG.info("torch found.")
except (ModuleNotFoundError, ImportError):
    LOG.info("torch not found.")
    th = None


CUPY_MODULES = {
    "scipy": "cupyx.scipy",
    "skimage": "cucim.skimage",
}


def is_cupy_array(arr: ArrayLike) -> bool:
    """
    Checks if array is a cupy array.

    Parameters
    ----------
    arr : ArrayLike
        Array to be checked.
    """
    return cp is not None and isinstance(arr, cp.ndarray)


@contextmanager
def unified_memory() -> Generator:
    """
    Initializes cupy's unified memory allocation.
    cupy's functions will run slower but it will spill memory memory into cpu without crashing.
    """
    # starts unified memory
    previous_allocator = None
    if cp is not None:
        previous_allocator = cp.cuda.get_allocator()
        cp.cuda.set_allocator(malloc_managed)

    yield

    # ends unified memory
    if cp is not None:
        cp.clear_memo()
        cp.cuda.set_allocator(previous_allocator)


def torch_default_device() -> "th.device":
    """
    Returns "gpu", "mps" or "cpu" devices depending on their availability.

    Returns
    -------
    th.device
        Torch fastest device.
    """
    if th is None:
        raise ImportError("torch not found.")

    if th.cuda.is_available():
        device = th.cuda.device_count() - 1
    elif th.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
    return th.device(device)


def import_module(
    module: str,
    submodule: str,
    arr: Optional[ArrayLike] = None,
) -> ModuleType:
    """Import GPU accelerated module if available and matches optional array, otherwise returns CPU version.

    Parameters
    ----------
    module : str
        Main python module (e.g. skimage, scipy)
    submodule : str
        Secondary python module (e.g. morphology, ndimage)
    arr : ArrayLike
        If provided, it will be used to determine if GPU module should be imported.

    Returns
    -------
    ModuleType
        Imported submodule.
    """
    is_gpu_array = cp is not None and isinstance(arr, cp.ndarray)

    if arr is None or is_gpu_array:
        cupy_module_name = f"{CUPY_MODULES[module]}.{submodule}"
        try:
            pkg = importlib.import_module(cupy_module_name)
            LOG.info(f"{cupy_module_name} found.")
            return pkg

        except (ModuleNotFoundError, ImportError):
            LOG.info(f"{cupy_module_name} not found. Using cpu equivalent")

    pkg = importlib.import_module(f"{module}.{submodule}")

    return pkg


def to_cpu(arr: ArrayLike) -> ArrayLike:
    """Moves array to cpu, if it's already there nothing is done."""
    if hasattr(arr, "cpu"):
        arr = arr.cpu()
    elif hasattr(arr, "get"):
        arr = arr.get()
    return arr


def on_gpu(func: Callable) -> Callable:
    """Decorator to run a function on GPU if available, otherwise it will run on CPU.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns
    -------
    Callable
        Decorated function.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if cp is not None:
            args = [
                cp.asarray(a) if isinstance(a, (np.ndarray, da.Array)) else a
                for a in args
            ]
            kwargs = {
                k: cp.asarray(v) if isinstance(v, (np.ndarray, da.Array)) else v
                for k, v in kwargs.items()
            }
        return func(*args, **kwargs)

    if not hasattr(func, "__name__"):
        # if it instead a class
        wrapper.__name__ = func.__class__.__name__

    return wrapper



import operator
from typing import Tuple

import numpy as np
import pytest

from ultrack.config import MainConfig
from ultrack.core.database import NodeDB
from ultrack.utils.array import array_apply
from ultrack.utils.ultrack_array import UltrackArray


@pytest.mark.parametrize("axis", [0, 1])
def test_array_apply_parametrized(axis):
    # Create sample data
    in_data = np.array([[1, 2, 3], [4, 5, 6]])
    out_data = np.zeros_like(in_data)

    # Define a sample function to apply
    def sample_func(arr_1, arr_2):
        return arr_1 + arr_2 + len(arr_1)

    array_apply(in_data, in_data, out_array=out_data, func=sample_func, axis=axis)
    other_axes_length = in_data.shape[1 - axis]
    assert np.array_equal(out_data, 2 * in_data + other_axes_length)


@pytest.mark.parametrize(
    "key,timelapse_mock_data",
    [
        (1, {"n_dim": 3}),
        (1, {"n_dim": 2}),
        ((slice(None), 1), {"n_dim": 3}),
        ((slice(None), 1), {"n_dim": 2}),
        ((0, [1, 2]), {"n_dim": 3}),
        ((0, [1, 2]), {"n_dim": 2}),
        # ((-1, np.asarray([0, 3])),{'n_dim':3}),       #does testing negative time make sense?
        # ((-1, np.asarray([0, 3])),{'n_dim':2}),
        ((slice(1), -2), {"n_dim": 3}),
        ((slice(1), -2), {"n_dim": 2}),
        ((np.asarray(0),), {"n_dim": 3}),
        ((np.asarray(0),), {"n_dim": 2}),
        ((0, 0, slice(32)), {"n_dim": 3}),
        ((0, 0, slice(32)), {"n_dim": 2}),
    ],
    indirect=[
        "timelapse_mock_data",
    ],
)
def test_ultrack_array(
    segmentation_database_mock_data: MainConfig,
    key: Tuple,
):
    ua = UltrackArray(segmentation_database_mock_data)
    ua_numpy = ua[slice(None)]
    np.testing.assert_equal(ua_numpy[key], ua[key])


def test_ultrack_array_filters(segmentation_database_mock_data: MainConfig):
    """Test that different filters work correctly."""

    # Test default filter (area <= threshold)
    ua_default = UltrackArray(segmentation_database_mock_data)

    # Test custom filter with id % 3 != 1
    ua_id_mod_3 = UltrackArray(
        segmentation_database_mock_data,
        filter_attr=NodeDB.id,
        filter_op=operator.mod,
        filter_value=3,
    )

    # Test custom filter with id % 2 != 1
    ua_id_mod_2 = UltrackArray(
        segmentation_database_mock_data,
        filter_attr=NodeDB.id,
        filter_op=operator.mod,
        filter_value=2,
    )

    # Test that different filters produce different results
    # (assuming the mock data has some variety)
    result_default = ua_default[0]
    result_selected = ua_id_mod_3[0]
    result_large = ua_id_mod_2[0]

    assert isinstance(result_default, np.ndarray)
    assert isinstance(result_selected, np.ndarray)
    assert isinstance(result_large, np.ndarray)

    # Check if non-empty
    assert np.any(result_default)
    assert np.any(result_selected)
    assert np.any(result_large)

    # At least one of these should be different from the others
    # (this is a basic sanity check)
    assert np.any(result_default != result_selected)
    assert np.any(result_default != result_large)
    assert np.any(result_selected != result_large)


def test_ultrack_array_set_filter(segmentation_database_mock_data: MainConfig):
    """Test that set_filter method works correctly."""

    ua = UltrackArray(segmentation_database_mock_data)
    original_result = ua[0].copy()

    # Change the filter
    ua.set_filter(NodeDB.selected, operator.eq, True)
    new_result = ua[0]

    # Results should be different (assuming mock data has variety)
    assert isinstance(original_result, np.ndarray)
    assert isinstance(new_result, np.ndarray)

    assert np.any(original_result != new_result)



from typing import Callable, Literal

import higra as hg
from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
    field_validator,
    model_validator,
)

NAME_TO_WS_HIER = {
    "area": hg.watershed_hierarchy_by_area,
    "dynamics": hg.watershed_hierarchy_by_dynamics,
    "volume": hg.watershed_hierarchy_by_volume,
}


class SegmentationConfig(BaseModel):
    """Segmentation hypotheses creation configuration"""

    min_area: int = 100
    """
    Minimum number of pixels for segmentation hypotheses prunning.
    Regions smaller than this value are merged or removed when there is no neighboring region
    """

    min_area_factor: float = 4.0
    """
    Objects in the foreground below `min_area / min_area_factor` are removed
    and not considered as hypotheses.
    """

    max_area: int = 1_000_000
    """Maximum number of pixels for segmentation hypotheses prunning
    Regions larger than this value are merged or removed when there is no neighboring region
    """

    n_workers: int = 1
    """Number of worker threads """

    min_frontier: float = 0.0
    """
    Minimum average frontier value between candidate segmentations, regions sharing an average
    frontier value lower than this are merged
    """

    threshold: float = 0.5
    """Threshold used to binarize the cell foreground map"""

    max_noise: float = 0.0
    """``SPECIAL``: Upper limit of uniform distribution for additive noise on contour map """

    random_seed: Literal["frame", "none", None] = "frame"
    """``SPECIAL``: Random seed initialization, if `frame` the seed is the timelapse frame number """

    ws_hierarchy: Callable = hg.watershed_hierarchy_by_area
    """
    ``SPECIAL``: Watershed hierarchy function from
    `higra <https://higra.readthedocs.io/en/stable/python/watershed_hierarchy.html>`_
    used to construct the hierarchy
    """

    anisotropy_penalization: float = 0.0
    """
    ``SPECIAL``: Image graph z-axis penalization, positive values will prioritize segmenting
    the xy-plane first, negative will do the opposite
    """

    model_config = ConfigDict(
        use_enum_values=True,
        extra="forbid",
    )

    @field_validator("ws_hierarchy", mode="before")
    def ws_name_to_function(cls, value: str) -> Callable:
        """Converts string to watershed hierarchy function."""
        if not isinstance(value, str):
            raise ValueError(
                f"`ws_hierarchy` must be a string. Found {value} of type {type(value)}."
            )

        if value not in NAME_TO_WS_HIER:
            raise ValueError(
                f"`ws_hierarchy` must match {NAME_TO_WS_HIER.keys()}. Found {value}"
            )

        return NAME_TO_WS_HIER[value]

    @field_serializer("ws_hierarchy")
    def serialize_ws_hierarchy(self, value: Callable) -> str:
        for name, func in NAME_TO_WS_HIER.items():
            if func == value:
                return name
        raise ValueError(
            f"ws_hierarchy function {value} not found in {NAME_TO_WS_HIER.keys()}"
        )

    @model_validator(mode="after")
    def area_validator(self) -> "SegmentationConfig":
        """Checks if min and max area are compatible."""
        if self.min_area > self.max_area:
            raise ValueError(
                "`min_area` must be lower than `max_area`."
                f"Found min_area={self.min_area}, max_area={self.max_area}"
            )
        return self



import os
from multiprocessing import Process
from pathlib import Path
from typing import Union

import uvicorn

from ultrack import MainConfig
from ultrack.api import app


def _in_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def start_server(
    api_results_path: Union[Path, str, None] = None,
    ultrack_data_config: Union[MainConfig, None] = None,
    host: str = "0.0.0.0",
    port: int = 8000,
):
    """Starts the server.

    Parameters
    ----------
    api_results_path : Union[Path, str, None], optional
        Path to the API results folder. Defaults to None.

    """
    if api_results_path is not None:
        os.environ["API_RESULTS_PATH"] = str(api_results_path)
    if ultrack_data_config is not None:
        os.environ["ULTRACK_DATA_CONFIG"] = ultrack_data_config.json()

    if _in_notebook():

        def start_in_notebook():
            uvicorn.run(app.app, host=host, port=port)

        Process(target=start_in_notebook).start()
    else:
        uvicorn.run(app.app, host=host, port=port)


if __name__ == "__main__":
    start_server()



import asyncio
import json
import logging
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import zarr
from fastapi import FastAPI, HTTPException, WebSocket
from napari.components import ViewerModel
from napari.plugins import _initialize_plugins
from networkx.readwrite import json_graph
from numpy._typing import ArrayLike
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import JSONResponse
from starlette.websockets import WebSocketState

import ultrack
from ultrack import link, segment, solve, to_tracks_layer, tracks_to_zarr
from ultrack.api.database import (
    Experiment,
    ExperimentStatus,
    create_experiment_instance,
    get_experiment,
    update_experiment,
)
from ultrack.api.settings import settings
from ultrack.api.utils.api import UltrackWebsocketLogger, raise_error
from ultrack.api.utils.fs import open_image
from ultrack.api.utils.zarr import get_channels_from_ome_zarr
from ultrack.config import MainConfig
from ultrack.core.export import tracks_layer_to_networkx, tracks_layer_to_trackmate
from ultrack.imgproc import detect_foreground, robust_invert
from ultrack.utils import labels_to_contours
from ultrack.utils.array import array_apply, create_zarr
from ultrack.utils.cuda import on_gpu

LOG = logging.getLogger(__name__)


app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["null", "*", "http://localhost:3000", "http://localhost:8080"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["*"],
)

app.state.queue = asyncio.Queue()


@app.exception_handler(Exception)
async def exception_handler(_, exc: Exception) -> JSONResponse:
    """General FastAPI exception handler.

    It captures all untreated exceptions and avoid the server to crash, returning a
    JSONResponse with the exception details. It also removes the experiment from the
    execution queue if it is still there.

    Parameters
    ----------
    _ : Request
        The request that raised the exception. it is not used.

    exc : Exception
        The exception raised.

    Returns
    -------
    JSONResponse
        A JSONResponse with the exception details.
    """
    LOG.error(exc)
    if app.state.queue.qsize() > 0:
        app.state.queue.task_done()
    return JSONResponse(
        status_code=exc.status_code,
        content={"detail": f"HTTPException: {exc.detail}"},
    )


def read_images_from_experiment(
    experiment: Experiment, image_names: List[str]
) -> dict[str, ArrayLike]:
    image_mapping = {
        "labels": experiment.labels_channel_or_path,
        "detection": experiment.detection_channel_or_path,
        "edges": experiment.edges_channel_or_path,
        "image": experiment.image_channel_or_path,
    }
    selected = [(im_name, image_mapping[im_name]) for im_name in image_names]

    if experiment.data_url is not None:
        node = open_image(experiment.data_url)
        named_data = get_channels_from_ome_zarr(node, valid_channels=selected)
    else:
        _initialize_plugins()
        viewer = ViewerModel()
        named_data = {
            name: viewer.open(path=path, plugin="napari")[0].data
            for name, path in selected
        }
    return named_data


async def start_experiment(ws: WebSocket, exp: Experiment) -> None:
    """Start an experiment.

    It persists the experiment instance to the database, updates its status to QUEUED,
    waits for the queue to be available, puts the experiment in the queue, and sends
    the experiment details to the client through the WebSocket.

    Parameters
    ----------
    ws : WebSocket
        The WebSocket instance used to communicate with the client.
    exp : Experiment
        The experiment instance with the ultrack configurations to be started.

    See Also
    --------
    finish_experiment : Finish the started experiment.
    """
    create_experiment_instance(exp)
    exp.status = ExperimentStatus.QUEUED
    update_experiment(exp)
    await app.state.queue.join()
    await app.state.queue.put(exp)
    await ws.send_json(json.loads(exp.model_dump_json()))
    UltrackWebsocketLogger.register_interruption_handler(exp.id)
    exp.status = ExperimentStatus.INITIALIZING
    update_experiment(exp)


async def finish_experiment(ws: WebSocket, exp: Experiment) -> Experiment:
    """Finish an experiment.

    Updates the experiment end time, and sends the final experiment state to the client
    through the WebSocket if it is still connected. It also removes the experiment from
    the execution queue.

    Parameters
    ----------
    ws : WebSocket
        The WebSocket instance used to communicate with the client.
    exp : Experiment
        The experiment instance to be finished.

    """
    exp.end_time = datetime.now().isoformat()
    exp.status = ExperimentStatus.SUCCESS
    update_experiment(exp)
    app.state.queue.task_done()
    if ws.client_state == WebSocketState.CONNECTED:
        # send the final state
        await ws.send_json(json.loads(exp.model_dump_json()))
    return exp


@app.get("/")
async def root() -> Dict[str, str]:
    """Root endpoint.

    Returns
    -------
    Dict[str, str]
        A dictionary with the ultrack version.
    """
    return {"ultrack.__version__": ultrack.__version__}


@app.get("/config/default")
async def get_default_config() -> Dict:
    """Gets the default ultrack configuration.

    Returns
    -------
    MainConfig
        The default ultrack configuration.
    """
    config = MainConfig()
    config.data_config = None
    return config.model_dump(by_alias=True)


@app.get("/config/available")
async def get_available_configs() -> Dict:
    """Gets the default ultrack configuration.

    Returns
    -------
    MainConfig
        The default ultrack configuration.
    """
    default_config = MainConfig()
    default_config.data_config = None

    experiment = {
        "name": "Unnamed Experiment",
        "config": default_config.model_dump(by_alias=True),
    }

    auto_detect_config = {
        "experiment": experiment.copy(),
        "detect_foreground_kwargs": {},
        "robust_invert_kwargs": {},
    }
    auto_detect_config["experiment"]["image_channel_or_path"] = "..."

    manual_segment_config = {
        "experiment": experiment.copy(),
    }
    manual_segment_config["experiment"]["edges_channel_or_path"] = "..."
    manual_segment_config["experiment"]["detection_channel_or_path"] = "..."

    labels_to_edges_config = {
        "experiment": experiment.copy(),
        "label_to_edges_kwargs": {},
    }
    labels_to_edges_config["experiment"]["labels_channel_or_path"] = "..."

    return {
        "auto_detect": {
            "link": "/segment/auto_detect",
            "human_name": "Auto Detect Foreground and Edges From Image",
            "config": auto_detect_config,
        },
        "manual_segment": {
            "link": "/segment/manual",
            "config": manual_segment_config,
            "human_name": "Manual Input of Foreground and Edges",
        },
        "labels_to_edges": {
            "link": "/segment/labels",
            "config": labels_to_edges_config,
            "human_name": "Auto Detect Foreground and Edges from Labels",
        },
    }


@app.post("/stop/{experiment_id}")
async def request_experiment_interruption(experiment_id: int) -> Dict[str, str]:
    """Request the interruption of an experiment given its id.

    Parameters
    ----------
    experiment_id : int
        The id of the experiment to be interrupted.

    Returns
    -------
    Dict[str, str]
        A dictionary with the status of the interruption request.
    """
    UltrackWebsocketLogger.request_interruption(experiment_id)
    return {"status": "ok"}


@app.websocket("/segment/auto_detect")
async def auto_detect(websocket: WebSocket) -> None:
    """Detect automatically foreground and boundaries by simple image processing.

    The experiment is only started if the WebSocket is connected and the experiment is
    provided. See the Notes section for the expected JSON structure.


    Parameters
    ----------
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.

    Notes
    -----
    The socket should receive a JSON with the following structure:
    {
        "experiment": {
            "data_url": "path/to/image.ome.zarr",
            "name": "Experiment Name",
            "image_channel_or_path": "ome_zarr channel_name",
            config: {
                ...
            },
        },
        "detect_foreground_kwargs": {
            ...
        },
        "robust_invert_kwargs": {
            ...
        }
    }
    where "detect_foreground_kwargs" and "robust_invert_kwargs" are optional and are
    the keyword arguments to be passed to the detect_foreground and robust_invert
    functions, respectively. experiment.config is a dictionary with the ultrack
    configuration as described in `ultrack.config.config.MainConfig`.

    See Also
    --------
    ultrack.imgproc.segmentation.detect_foreground : Automatically detects foreground
        given an image.
    ultrack.imgproc.segmentation.robust_invert : Automatically detects edges given an
        image.
    ultrack.api.database.Experiment : Experiment class definition.
    ultrack.config.config.MainConfig : ultrack configuration class definition.
    """
    await websocket.accept()
    data = await websocket.receive_json()

    if "experiment" not in data:
        await raise_error(
            ValueError("Experiment not provided. Please provide an experiment."),
            ws=websocket,
        )
        return

    experiment = Experiment.model_validate(data["experiment"])

    try:
        detect_foreground_kwargs = data["detect_foreground_kwargs"]
    except KeyError:
        detect_foreground_kwargs = {}
        LOG.warning("detect_foreground_kwargs not provided. Using default values.")

    try:
        robust_invert_kwargs = data["robust_invert_kwargs"]
    except KeyError:
        robust_invert_kwargs = {}
        LOG.warning("robust_invert_kwargs not provided. Using default values.")

    async with UltrackWebsocketLogger(websocket, experiment):
        await start_experiment(websocket, experiment)

        named_data = read_images_from_experiment(experiment, image_names=["image"])

        image_data: zarr.Array = named_data["image"]

        shape = image_data.shape

        with tempfile.TemporaryDirectory() as temp_path:
            detection_path = str(Path(temp_path) / "detection.zarr")
            edges_path = str(Path(temp_path) / "edges.zarr")

            zarr_detection = create_zarr(
                shape=shape,
                dtype=float,
                store_or_path=detection_path,
                overwrite=True,
            )
            array_apply(
                image_data.astype("float32"),
                out_array=zarr_detection,
                func=on_gpu(detect_foreground),
                **detect_foreground_kwargs,
            )

            zarr_edges = create_zarr(
                shape=shape,
                dtype=float,
                store_or_path=edges_path,
                overwrite=True,
            )
            array_apply(
                image_data.astype("float32"),
                out_array=zarr_edges,
                func=on_gpu(robust_invert),
                **robust_invert_kwargs,
            )

            experiment.data_url = None
            experiment.image_channel_or_path = None
            experiment.labels_channel_or_path = None
            experiment.edges_channel_or_path = edges_path
            experiment.detection_channel_or_path = detection_path
            experiment.status = ExperimentStatus.DATA_LOADED

            update_experiment(experiment)

            await segment_link_and_solve(experiment, ws=websocket)

    await finish_experiment(websocket, experiment)


@app.websocket("/segment/manual")
async def manual_segment(websocket: WebSocket) -> None:
    """Manually provide the foreground and edges, subsequently segmenting and solving
    the experiment by ultrack.

    The experiment is only started if the WebSocket is connected and the experiment is
    provided. See the Notes section for the expected JSON structure.

    Parameters
    ----------
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.

    Notes
    -----
    The socket should receive a JSON with the following structure:
    {
        "experiment": {
            "data_url": "path/to/image.ome.zarr",
            "name": "Experiment Name",
            "detection_channel": "ome_zarr_channel_name",
            "edges_channel": "ome_zarr_edges_channel_name",
            config: {
                ...
            },
        }
    }
    experiment.config is a dictionary with the ultrack
    configuration as described in `ultrack.config.config.MainConfig`.

    See Also
    --------
    ultrack.api.database.Experiment : Experiment class definition.
    ultrack.config.config.MainConfig : ultrack configuration class definition.
    """
    await websocket.accept()
    data = await websocket.receive_json()

    if "experiment" not in data:
        await raise_error(
            ValueError("Experiment not provided. Please provide an experiment."),
            ws=websocket,
        )
        return

    if isinstance(data["experiment"], str):
        experiment = json.loads(data["experiment"])
    else:
        experiment = data["experiment"]
    experiment = Experiment.model_validate(experiment)

    async with UltrackWebsocketLogger(websocket, experiment):
        await start_experiment(websocket, experiment)

        experiment.status = ExperimentStatus.DATA_LOADED
        update_experiment(experiment)

        await segment_link_and_solve(experiment, websocket)

    await finish_experiment(websocket, experiment)


@app.websocket("/segment/labels")
async def auto_from_labels(websocket: WebSocket) -> None:
    """Detect automatically foreground and boundaries by simple image processing from
    an external label map provided by an instance segmentation algorithm or model.

    The experiment is only started if the WebSocket is connected and the experiment is
    provided. See the Notes section for the expected JSON structure.

    Parameters
    ----------
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.

    Notes
    -----
    The socket should receive a JSON with the following structure:
    {
        "experiment": {
            "data_url": "path/to/image.ome.zarr",
            "name": "Experiment Name",
            "label_channel": "ome_zarr_label_channel_name",
            config: {
                ...
            },
        },
        "label_to_edges_kwargs": {
            ...
        },
    }
    where "label_to_edges_kwargs" is optional and is the keyword arguments to be passed
    to the labels_to_edges function. experiment.config is a dictionary with the ultrack
    configuration as described in `ultrack.config.config.MainConfig`.

    See Also
    --------
    ultrack.imgproc.segmentation.labels_to_edges : Automatically detects foreground
        and edges given a label map.
    ultrack.api.database.Experiment : Experiment class definition.
    ultrack.config.config.MainConfig : ultrack configuration class definition.
    """
    await websocket.accept()
    data = await websocket.receive_json()

    if "experiment" not in data:
        await raise_error(
            ValueError("Experiment not provided. Please provide an experiment."),
            ws=websocket,
        )
        return

    experiment = Experiment.model_validate(data["experiment"])

    try:
        label_to_edges_kwargs = data["label_to_edges_kwargs"]
    except KeyError:
        label_to_edges_kwargs = {}
        LOG.warning("label_to_edges_kwargs not provided. Using default values.")

    async with UltrackWebsocketLogger(websocket, experiment):
        await start_experiment(websocket, experiment)

        named_data = read_images_from_experiment(experiment, image_names=["labels"])

        label_data: zarr.Array = named_data["labels"]

        with tempfile.TemporaryDirectory() as temp_path:
            detection_path = str(Path(temp_path) / "detection.zarr")
            edges_path = str(Path(temp_path) / "edges.zarr")

            labels_to_contours(
                label_data,
                foreground_store_or_path=detection_path,
                contours_store_or_path=edges_path,
                **label_to_edges_kwargs,
            )

            experiment.data_url = None
            experiment.image_channel_or_path = None
            experiment.labels_channel_or_path = None
            experiment.edges_channel_or_path = edges_path
            experiment.detection_channel_or_path = detection_path
            experiment.status = ExperimentStatus.DATA_LOADED

            update_experiment(experiment)

            await segment_link_and_solve(experiment, websocket)

    await finish_experiment(websocket, experiment)


async def segment_link_and_solve(experiment: Experiment, ws: WebSocket) -> None:
    """Segment, link, and solve an experiment with Ultrack.

    Parameters
    ----------
    experiment : Experiment
        The experiment to be executed. It should have the data_url, detection_channel,
        and edges_channel set.

    ws : WebSocket
        The WebSocket instance used to communicate with the client.

    Raises
    ------
    ProcessedException
        If any error occurs during the execution of the experiment.
    """
    try:
        named_data = read_images_from_experiment(
            experiment, image_names=["detection", "edges"]
        )

        detection = named_data["detection"]
        edges = named_data["edges"]

        config = experiment.get_config()
        experiment.status = ExperimentStatus.SEGMENTING
        update_experiment(experiment)
        segment(
            foreground=detection,
            contours=edges,
            config=config,
        )
    except Exception as e:
        await raise_error(e, experiment, ws)

    experiment.status = ExperimentStatus.LINKING
    update_experiment(experiment)
    link(config)

    experiment.status = ExperimentStatus.SOLVING
    update_experiment(experiment)
    solve(config)

    try:
        experiment.status = ExperimentStatus.EXPORTING
        update_experiment(experiment)
        tracks_df, graph = to_tracks_layer(config)
        G = tracks_layer_to_networkx(tracks_df)
        G = json_graph.node_link_data(G)
        experiment.tracks = json.dumps(G)
        segments_path = (
            settings.api_results_path / f"segments_id_exp_{experiment.id:06d}.zarr"
        )
        tracks_to_zarr(
            config,
            tracks_df,
            store_or_path=segments_path,
            overwrite=True,
        )
        experiment.final_segments_url = str(segments_path.absolute())
    except Exception as e:
        await raise_error(
            e, experiment, ws, f"Could not export tracks to zarr. {str(e)}"
        )


@app.get("/experiment/{experiment_id}/trackmate")
async def get_trackmate(experiment_id: int) -> Dict:
    """Get the TrackMate xml from an experiment.

    Parameters
    ----------
    experiment_id : int
        The id of the experiment to get the TrackMate xml.

    Returns
    -------
    Dict
        A dictionary with the TrackMate XML file.
    """
    experiment = get_experiment(experiment_id)
    if experiment is None:
        raise HTTPException(
            status_code=404, detail=f"Experiment {experiment_id} not found."
        )
    tracks_df, _ = to_tracks_layer(experiment.get_config())
    trackmate_xml = tracks_layer_to_trackmate(tracks_df)
    return {"trackmate_xml": trackmate_xml}


@app.websocket("/segment/model/{model_name}")
async def segment_by_model(websocket: WebSocket) -> None:
    """TODO: Segment by a model.

    Available models: cellpose, stardist2d, PlantSeg, and MicroSAM.

    Parameters
    -----------
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.
    """
    raise HTTPException(status_code=501)


async def add_flow(experiment: Experiment, websocket: WebSocket) -> None:
    """TODO: Add flow to an experiment.

    Parameters
    -----------
    experiment : Experiment
        The experiment to be executed.
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.
    """
    raise HTTPException(status_code=501)


async def auto_flow(experiment: Experiment, websocket: WebSocket) -> None:
    """TODO: Automatically compute the flow of an experiment.

    Parameters
    -----------
    experiment : Experiment
        The experiment to be executed.
    websocket : WebSocket
        The WebSocket instance used to communicate with the client.
    """
    raise HTTPException(status_code=501)



import numpy as np

from ultrack.imgproc import normalize


def test_basic_normalization() -> None:
    image = np.array([[[[-0.2, 0.5], [0.4, 0.8]], [[0.6, 0.3], [1.1, 0.9]]]])
    normalized = normalize(image, 1.0)

    assert np.all(normalized >= 0) and np.all(normalized <= 1)
    assert not np.array_equal(image, normalized)



from enum import Enum
from typing import Callable, Literal, Optional, Tuple

import numpy as np
from pydantic import BaseModel, ConfigDict


class LinkFunctionChoices(Enum):
    identity = "identity"
    power = "power"


class TrackingConfig(BaseModel):
    """Tracking (segmentation & linking selection) configuration"""

    solver_name: Literal["GUROBI", "CBC", ""] = ""
    """
    Constrained optimization solver name.

    * GUROBI: Commercial solver, requires license, see :ref:`gurobi_install` for extra information.

    * CBC: Open-source solver, slower, uses more memory than Gurobi and harder to install on Window.

    * "": Use default solver, GUROBI if available, otherwise CBC.

    """

    appear_weight: float = -0.001
    """Penalization weight for appearing cell, should be negative """
    disappear_weight: float = -0.001
    """Penalization for disappearing cell, should be negative """
    division_weight: float = -0.001
    """Penalization for dividing cell, should be negative """
    image_border_size: Optional[Tuple[int, ...]] = None
    """Image border size in pixels (Z,Y,X) to avoid tracking cells within this border.
       If cells are within the border they not penalized when appearing or disappearing
    """
    n_threads: int = -1
    """Number of worker threads """

    window_size: Optional[int] = None
    """
    Time window size for partially solving the tracking ILP.
    By default it solves the entire timelapse at once.
    Useful for large datasets.
    """

    overlap_size: int = 1
    """
    Number of frames used to shared (overlap/pad) each size when ``window_size`` is set.
    This improves the tracking quality at the edges of the windows and enforce continuity of the tracks.
    """

    solution_gap: float = 0.001
    """
    Solver solution gap. This will speed up the solver when finding the optimal
    solution might taken a long time, but may affect the quality of the solution.
    """

    time_limit: int = 36000
    """Solver execution time limit in seconds """

    method: int = 0
    """``SPECIAL``: Solver method, `reference <https://docs.python-mip.com/en/latest/classes.html#lp-method>`_"""

    link_function: LinkFunctionChoices = "power"
    """``SPECIAL``: Function used to transform the edge and node weights, `identity` or `power`"""

    power: float = 4
    r"""``SPECIAL``: Expoent :math:`\eta` of power transform, :math:`w_{pq}^\eta` """

    bias: float = -0.0
    """``SPECIAL``: Edge weights bias :math:`b`, :math:`w_{pq} + b`, should be negative """

    dismiss_weight_guess: Optional[float] = None
    include_weight_guess: Optional[float] = None

    model_config = ConfigDict(
        use_enum_values=True,
        extra="forbid",
    )

    @property
    def apply_link_function(self) -> Callable[[np.ndarray], np.ndarray]:
        if self.link_function == "identity":
            return lambda x: x + self.bias
        elif self.link_function == "power":
            return lambda x: np.power(x, self.power) + self.bias
        else:
            raise NotImplementedError



from typing import List, Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike
from sqlalchemy import create_engine, func
from sqlalchemy.orm import Session

from ultrack.config.config import LinkingConfig, MainConfig
from ultrack.core.database import LinkDB, NodeDB, OverlapDB
from ultrack.core.segmentation.node import Node


def _nearest_neighbors(
    data_arr: ArrayLike,
    node: Node,
    n_neighbors: int,
    max_distance: Optional[float],
    scale: Optional[ArrayLike],
) -> np.ndarray:
    """
    Returns the indices of the `n_neighbors` nearest neighbors to `node` in `data_arr`.

    Parameters
    ----------
    data_arr : ArrayLike
        Array of (id, z, y, x) coordinates of the nodes (or centroids).
    node : Node
        Node to find neighbors for.
    n_neighbors : int
        Number of neighbors to be considered.
    max_distance : float
        Maximum distance to be considered.
    scale : Optional[ArrayLike], optional
        Scaling factor for the distance, by default None.

    Returns
    -------
    np.ndarray
        Indices of the nearest neighbors.
    """
    if scale is None:
        scale = np.ones(len(node.centroid))

    differences = data_arr[:, -len(node.centroid) :] - node.centroid
    differences *= scale
    sqdist = np.square(differences).sum(axis=1)

    if max_distance is not None:
        valid = sqdist <= (max_distance * max_distance)
        data_arr = data_arr[valid]
        sqdist = sqdist[valid]

    indices = np.argsort(sqdist)

    return data_arr[indices[:n_neighbors], 0]


def _find_links(
    session: Session,
    node: Node,
    adj_time: int,
    scale: Optional[ArrayLike],
    link_config: LinkingConfig,
) -> List[Tuple[NodeDB, float]]:
    """
    Finds links for a given node and time.

    Parameters
    ----------
    session : Session
        SQLAlchemy session.
    node : Node
        Node to search for neighbors.
    adj_time : int
        Adjacent time point.
    scale : Optional[ArrayLike], optional
        Scaling factor for the distance, by default None.
    link_config : LinkingConfig
        Linking configuration parameters.

    Returns
    -------
    List[NodeDB, float]
        List of nodes and their weights.
    """
    data = np.asarray(
        session.query(
            NodeDB.id,
            NodeDB.z,
            NodeDB.y,
            NodeDB.x,
        )
        .where(NodeDB.t == adj_time)
        .all()
    )
    if len(data) == 0:
        return []

    ind = _nearest_neighbors(
        data,
        node,
        2 * link_config.max_neighbors,
        link_config.max_distance,
        scale=scale,
    )

    neigh_nodes = session.query(NodeDB.pickle).where(NodeDB.id.in_(ind)).all()

    if scale is None:
        scale = np.ones(len(node.centroid))

    neigh_nodes_with_dist = []
    for (n,) in neigh_nodes:
        dist = np.linalg.norm((n.centroid - node.centroid) * scale)
        w = node.IoU(n) - link_config.distance_weight * dist
        neigh_nodes_with_dist.append((n, w))

    neigh_nodes_with_dist.sort(key=lambda x: x[1], reverse=True)

    return neigh_nodes_with_dist[: link_config.max_neighbors]


def _add_overlaps(
    session: Session,
    node: Node,
    n_neighbors: int = 10,
    scale: Optional[ArrayLike] = None,
) -> None:
    """
    Adds overlaps to the database.

    Parameters
    ----------
    session : Session
        SQLAlchemy session.
    node : Node
        Node to find overlaps with.
    n_neighbors : int, optional
        Number of neighbors to be considered, by default 10.
    scale : Optional[ArrayLike], optional
        Scaling factor for the distance, by default None.
    """
    data = np.asarray(
        session.query(
            NodeDB.id,
            NodeDB.z,
            NodeDB.y,
            NodeDB.x,
        )
        .where(NodeDB.t == node.time, NodeDB.id != node.id)
        .all()
    )
    ind = _nearest_neighbors(data, node, n_neighbors, max_distance=None, scale=scale)

    overlaps = []

    for (neigh_node,) in session.query(NodeDB.pickle).where(NodeDB.id.in_(ind)).all():
        if node.IoU(neigh_node) > 0.0:
            overlaps.append(
                OverlapDB(
                    node_id=node.id,
                    ancestor_id=neigh_node.id,
                )
            )

    session.add_all(overlaps)


def _add_links(
    session: Session,
    node: Node,
    link_config: LinkingConfig,
    scale: Optional[ArrayLike] = None,
) -> None:
    """
    Adds T - 1 and T + 1 links to the database.

    NOTE: this is not taking node shifts into account.

    Parameters
    ----------
    session : Session
        SQLAlchemy session.
    node : Node
        Node to search for neighbors.
    link_config : LinkingConfig
        Linking configuration parameters.
    scale : Optional[ArrayLike], optional
        Scaling factor for the distance, by default None.
    """
    links = []

    before_links = _find_links(
        session=session,
        node=node,
        adj_time=node.time - 1,
        scale=scale,
        link_config=link_config,
    )
    for before_node, w in before_links:
        links.append(
            LinkDB(
                source_id=before_node.id,
                target_id=node.id,
                weight=w,
            )
        )

    after_links = _find_links(
        session=session,
        node=node,
        adj_time=node.time + 1,
        scale=scale,
        link_config=link_config,
    )
    for after_node, w in after_links:
        links.append(
            LinkDB(
                source_id=node.id,
                target_id=after_node.id,
                weight=w,
            )
        )

    session.add_all(links)


def add_new_node(
    config: MainConfig,
    time: int,
    mask: ArrayLike,
    bbox: Optional[ArrayLike] = None,
    index: Optional[int] = None,
    include_overlaps: bool = True,
) -> int:
    """
    Adds a new node to the database.

    NOTE: this is not taking node shifts or image features (color) into account.

    Parameters
    ----------
    config : MainConfig
        Ultrack configuration parameters.
    time : int
        Time point of the node.
    mask : ArrayLike
        Binary mask of the node.
    bbox : Optional[ArrayLike], optional
        Bounding box of the node, (min_0, min_1, ..., max_0, max_1, ...).
        When provided it assumes the mask is a crop of the original image, by default None
    index : Optional[int], optional
        Node index, otherwise it is automatically generated, and returned.
    include_overlaps : bool, optional
        Include overlaps in the database, by default True
        When False it will allow oclusions between new node and existing nodes.

    Returns
    -------
    int
        New node index.
    """

    node = Node.from_mask(
        time=time,
        mask=mask,
        bbox=np.asarray(bbox),
    )
    if node.area == 0:
        raise ValueError("Node area is zero. Something went wrong.")

    scale = config.data_config.metadata.get("scale")

    engine = create_engine(config.data_config.database_path)
    with Session(engine) as session:

        # querying required data
        if index is None:
            node.id = (
                int(session.query(func.max(NodeDB.id)).where(NodeDB.t == time).scalar())
                + 1
            )
        else:
            node.id = index

        # adding node
        if len(node.centroid) == 2:
            y, x = node.centroid
            z = 0
        else:
            z, y, x = node.centroid

        node_db_obj = NodeDB(
            id=node.id,
            t=node.time,
            z=z,
            y=y,
            x=x,
            area=node.area,
            pickle=node,
        )
        session.add(node_db_obj)

        if include_overlaps:
            _add_overlaps(session=session, node=node, scale=scale)

        _add_links(
            session=session,
            node=node,
            link_config=config.linking_config,
            scale=scale,
        )

        session.commit()

    return node.id



import logging
from contextlib import nullcontext
from typing import Callable, List, Optional, Sequence

import fasteners
import numpy as np
import pandas as pd
import sqlalchemy as sqla
from numpy.typing import ArrayLike
from scipy.spatial import KDTree
from sqlalchemy.orm import Session
from toolz import curry

from ultrack.config.config import LinkingConfig, MainConfig
from ultrack.core.database import LinkDB, NodeDB, maximum_time_from_database
from ultrack.core.linking.utils import clear_linking_data
from ultrack.core.segmentation.node import Node
from ultrack.utils.array import check_array_chunk
from ultrack.utils.multiprocessing import (
    batch_index_range,
    multiprocessing_apply,
    multiprocessing_sqlite_lock,
)

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


def _compute_features(
    time: int,
    nodes: List[Node],
    images: Sequence[ArrayLike],
    feature_funcs: List[Callable[[Node, ArrayLike], ArrayLike]],
) -> List[ArrayLike]:
    """Compute mean intensity for each node"""

    frames = np.stack(
        [np.asarray(image[time]) for image in images],
        axis=-1,
    )
    LOG.info(f"Image with shape {[f.shape for f in frames]}")

    return [
        np.stack([func(node, frames) for node in nodes], axis=0)
        for func in feature_funcs
    ]


def color_filtering_mask(
    time: int,
    current_nodes: List[Node],
    next_nodes: List[Node],
    images: Sequence[ArrayLike],
    neighbors: ArrayLike,
    z_score_threshold: float,
) -> ArrayLike:
    """
    Filtering by color z-score.

    Parameters
    ----------
    time : int
        Current time.
    current_nodes : List[Node]
        List of source nodes.
    next_nodes : List[Node]
        List of target nodes.
    images : Sequence[ArrayLike]
        Sequence of images to extract color features for filtering.
    neighbors : ArrayLike
        Neighbors indices (current/source) for each target (next) node.
    z_score_threshold : float
        Z-score threshold for color filtering.

    Returns
    -------
    ArrayLike
        Boolean mask of neighboring nodes within color z-score threshold.

    """
    LOG.info(f"computing filtering by color z-score from t={time}")
    (current_features,) = _compute_features(
        time, current_nodes, images, [Node.intensity_mean]
    )
    # inserting dummy value for missing neighbors
    current_features = np.append(
        current_features,
        np.zeros((1, current_features.shape[1])),
        axis=0,
    )
    next_features, next_features_std = _compute_features(
        time + 1, next_nodes, images, [Node.intensity_mean, Node.intensity_std]
    )
    LOG.info(
        f"Features Std. Dev. range {next_features_std.min()} {next_features_std.max()}"
    )
    next_features_std[next_features_std <= 1e-6] = 1.0
    difference = next_features[:, None, ...] - current_features[neighbors]
    difference /= next_features_std[:, None, ...]
    filtered_by_color = np.abs(difference).max(axis=-1) <= z_score_threshold
    return filtered_by_color


@curry
def _process(
    time: int,
    config: LinkingConfig,
    db_path: str,
    images: Sequence[ArrayLike],
    scale: Optional[Sequence[float]],
    write_lock: Optional[fasteners.InterProcessLock] = None,
) -> None:
    """Link nodes from current time to time + 1.

    Parameters
    ----------
    time : int
        Current time.
    config : LinkingConfig
        Linking configuration parameters.
    db_path : str
        Database path.
    images : Sequence[ArrayLike]
        Sequence of images for color space filtering, if empty, no filtering is performed.
    scale : Sequence[float]
        Optional scaling for nodes' distances.
    write_lock : Optional[fasteners.InterProcessLock], optional
        Lock object for SQLite multiprocessing, optional otherwise, by default None.
    """
    connect_args = {"timeout": 45} if write_lock is not None else {}
    engine = sqla.create_engine(db_path, connect_args=connect_args)
    with Session(engine) as session:
        current_nodes = [
            n for n, in session.query(NodeDB.pickle).where(NodeDB.t == time)
        ]

        query = session.query(
            NodeDB.pickle,
            NodeDB.z_shift,
            NodeDB.y_shift,
            NodeDB.x_shift,
        ).where(NodeDB.t == time + 1)

        next_nodes = [row[0] for row in query]
        next_shift = np.asarray([row[1:] for row in query])

    compute_spatial_neighbors(
        time,
        config,
        current_nodes,
        next_nodes,
        next_shift,
        scale=scale,
        table_name=LinkDB.__tablename__,
        db_path=db_path,
        images=images,
        write_lock=write_lock,
    )


def compute_spatial_neighbors(
    time: int,
    config: LinkingConfig,
    source_nodes: List[Node],
    target_nodes: List[Node],
    target_shift: ArrayLike,
    scale: Optional[Sequence[float]],
    table_name: str,
    db_path: str,
    images: Sequence[ArrayLike],
    write_lock: Optional[fasteners.InterProcessLock] = None,
    weight_func: Callable[[Node, Node], float] = Node.IoU,
    edge_must_be_positive: bool = False,
) -> pd.DataFrame:

    source_pos = np.asarray([n.centroid for n in source_nodes])
    target_pos = np.asarray([n.centroid for n in target_nodes], dtype=np.float32)

    n_dim = target_pos.shape[1]
    target_shift = target_shift[:, -n_dim:]  # matching positions dimensions
    target_pos += target_shift

    if scale is not None:
        min_n_dim = min(n_dim, len(scale))
        scale = scale[-min_n_dim:]
        source_pos = source_pos[..., -min_n_dim:] * scale
        target_pos = target_pos[..., -min_n_dim:] * scale

    # finds neighbors nodes within the radius
    # and connect the pairs with highest edge weight
    current_kdtree = KDTree(source_pos)

    distances, neighbors = current_kdtree.query(
        target_pos,
        # twice as expected because we select the nearest with highest edge weight
        k=2 * config.max_neighbors,
        distance_upper_bound=config.max_distance,
    )

    if len(images) > 0:
        filtered_by_color = color_filtering_mask(
            time,
            source_nodes,
            target_nodes,
            images,
            neighbors,
            config.z_score_threshold,
        )
    else:
        filtered_by_color = np.ones_like(neighbors, dtype=bool)

    int_next_shift = np.round(target_shift).astype(int)
    # NOTE: moving bbox with shift, MUST be after `feature computation`
    for node, shift in zip(target_nodes, int_next_shift):
        node.bbox[:n_dim] += shift
        node.bbox[-n_dim:] += shift

    distance_w = config.distance_weight
    links = []

    for i, node in enumerate(target_nodes):
        valid = (~np.isinf(distances[i])) & filtered_by_color[i]
        valid_neighbors = neighbors[i, valid]
        neigh_distances = distances[i, valid]

        neighborhood = []
        for neigh_idx, neigh_dist in zip(valid_neighbors, neigh_distances):
            neigh = source_nodes[neigh_idx]
            edge_weight = weight_func(node, neigh) - distance_w * neigh_dist
            # using dist as a tie-breaker
            if edge_must_be_positive and edge_weight <= 0:
                continue
            neighborhood.append(
                (edge_weight, -neigh_dist, neigh.id, node.id)
            )  # current, next

        neighborhood = sorted(neighborhood, reverse=True)[: config.max_neighbors]
        LOG.info("Node %s links %s", node.id, neighborhood)
        links += neighborhood

    if len(links) == 0:
        raise ValueError(
            f"No links found for time {time}. Increase `linking_config.max_distance` parameter."
        )

    links = np.asarray(links)[:, [0, 2, 3]]  # ignoring index column
    df = pd.DataFrame(links, columns=["weight", "source_id", "target_id"])

    with write_lock if write_lock is not None else nullcontext():
        LOG.info(f"Pushing links from time {time} to {db_path}")
        connect_args = {"timeout": 45} if write_lock is not None else {}
        engine = sqla.create_engine(
            db_path, hide_parameters=True, connect_args=connect_args
        )
        with engine.begin() as conn:
            df.to_sql(name=table_name, con=conn, if_exists="append", index=False)

    return df


def link(
    config: MainConfig,
    images: Sequence[ArrayLike] = tuple(),
    scale: Optional[Sequence[float]] = None,
    batch_index: Optional[int] = None,
    overwrite: bool = False,
) -> None:
    """Links candidate segments (nodes) with their neighbors on the next time.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    images : Sequence[ArrayLike]
        Optinal sequence of images for color space filtering.
    scale : Sequence[float]
        Optional scaling for nodes' distances.
    batch_index : Optional[int], optional
        Batch index for processing a subset of nodes, by default everything is processed.
    overwrite : bool
        Cleans up linking database content before processing.
    """
    LOG.info(f"Linking nodes with LinkingConfig:\n{config.linking_config}")

    for image in images:
        check_array_chunk(image)

    max_t = maximum_time_from_database(config.data_config)
    time_points = batch_index_range(max_t, config.linking_config.n_workers, batch_index)
    LOG.info(f"Linking time points {time_points}")

    if overwrite and (batch_index is None or batch_index == 0):
        clear_linking_data(config.data_config.database_path)

    with multiprocessing_sqlite_lock(config.data_config) as lock:
        process = _process(
            config=config.linking_config,
            db_path=config.data_config.database_path,
            write_lock=lock,
            images=images,
            scale=scale,
        )
        multiprocessing_apply(
            process, time_points, config.linking_config.n_workers, desc="Linking nodes."
        )


def add_links(
    config: MainConfig,
    sources: ArrayLike,
    targets: ArrayLike,
    weights: ArrayLike,
) -> None:
    """
    Adds user-defined links to the database.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    sources : ArrayLike
        Sources (t) node id.
    targets : ArrayLike
        Targets (t + 1) node id.
    weights : ArrayLike
        Link weights, the higher the weight the more likely the link.
    """
    df = pd.DataFrame(
        {
            "source_id": np.asarray(sources, dtype=int),
            "target_id": np.asarray(targets, dtype=int),
            "weight": weights,
        }
    )

    engine = sqla.create_engine(
        config.data_config.database_path,
        hide_parameters=True,
    )

    with engine.begin() as conn:
        df.to_sql(name=LinkDB.__tablename__, con=conn, if_exists="append", index=False)



import numba
import numpy as np
from numba.experimental import jitclass

from ultrack.core.solve.solver.heuristic._heap import Heap, Policy

_SPEC = {
    "_objective": numba.float64,
    "_weights": numba.float32[:],
    "_in_edge": numba.int64[:],
    "_out_edge": numba.int64[:],
    "_forbidden": numba.boolean[:],
    "_in_count": numba.uint8[:],
    "_out_count": numba.uint8[:],
    "_selected_nodes": numba.boolean[:],
    "_predecessor_map": numba.int64[:],
    "_predecessor_weight": numba.float32[:],
    "_overlap_indices": numba.int32[:],
    "_overlap_indptr": numba.int32[:],
    "_in_out_digraph_indices": numba.int32[:],
    "_in_out_digraph_indptr": numba.int32[:],
    "_in_out_digraph_data": numba.float32[:],
    "_appear_weight": numba.float32[:],
    "_disappear_weight": numba.float32[:],
    "_division_weight": numba.float64,
    "_add_in_map": numba.float32[:, :, :],
    "_sub_in_map": numba.float32[:, :, :],
    "_add_out_map": numba.float32[:, :, :],
    "_sub_out_map": numba.float32[:, :, :],
}


@jitclass(_SPEC)
class NumbaHeuristicSolver:
    def __init__(
        self,
        weights: np.ndarray,
        in_edge: np.ndarray,
        out_edge: np.ndarray,
        forbidden: np.ndarray,
        in_count: np.ndarray,
        out_count: np.ndarray,
        selected_nodes: np.ndarray,
        predecessor_map: np.ndarray,
        predecessor_weight: np.ndarray,
        overlap_indices: np.ndarray,
        overlap_indptr: np.ndarray,
        in_out_digraph_indices: np.ndarray,
        in_out_digraph_indptr: np.ndarray,
        in_out_digraph_data: np.ndarray,
        appear_weight: np.ndarray,
        disappear_weight: np.ndarray,
        division_weight: float,
        add_in_map: np.ndarray,
        sub_in_map: np.ndarray,
        add_out_map: np.ndarray,
        sub_out_map: np.ndarray,
    ) -> None:

        self._objective = 0.0
        self._weights = weights
        self._in_edge = in_edge
        self._out_edge = out_edge
        self._forbidden = forbidden
        self._in_count = in_count
        self._out_count = out_count
        self._selected_nodes = selected_nodes
        self._predecessor_map = predecessor_map
        self._predecessor_weight = predecessor_weight
        self._overlap_indices = overlap_indices
        self._overlap_indptr = overlap_indptr
        self._in_out_digraph_indices = in_out_digraph_indices
        self._in_out_digraph_indptr = in_out_digraph_indptr
        self._in_out_digraph_data = in_out_digraph_data
        self._appear_weight = appear_weight
        self._disappear_weight = disappear_weight
        self._division_weight = division_weight
        self._add_in_map = add_in_map
        self._sub_in_map = sub_in_map
        self._add_out_map = add_out_map
        self._sub_out_map = sub_out_map

    def optimize(self) -> float:
        """Optimizes objective function."""
        np.random.seed(42)

        ascent_iters = 5
        obj_tol = 1
        self.constrained_mst()

        prev_objective = 0.0
        for _ in range(ascent_iters):
            self.random_ascent()
            if prev_objective - self._objective < obj_tol:
                break
            prev_objective = self._objective

        return self._objective

    def constrained_mst(self) -> None:
        """
        Executes Kruskal MST algorithm while preserving the biological constraints.
        """
        heap = Heap(self._weights, Policy.maximum)
        heap.insert_array(np.arange(len(self._weights)))

        while not heap.is_empty():
            edge_index = heap.pop()
            in_node = self._in_edge[edge_index]
            out_node = self._out_edge[edge_index]

            # blocked by overlap
            if self._forbidden[in_node] or self._forbidden[out_node]:
                continue

            # check for flow constrains
            if self._in_count[in_node] > 0 or self._out_count[out_node] > 1:
                continue

            weight = self._weights[edge_index]
            obj_delta = weight
            obj_delta += self._transition_weight(self._add_in_map, in_node)
            obj_delta += self._transition_weight(self._add_out_map, out_node)

            # any other check?
            self._objective += obj_delta

            self._selected_nodes[out_node] = True
            self._selected_nodes[in_node] = True
            self._predecessor_map[in_node] = out_node
            self._predecessor_weight[in_node] = weight

            self._out_count[out_node] += 1
            self._in_count[in_node] += 1
            self._forbid_overlap(in_node)
            self._forbid_overlap(out_node)

    def random_ascent(self) -> None:
        """Executes random ascent on the current solution."""
        n_nodes = len(self._appear_weight)
        node_ids = np.random.choice(n_nodes, size=n_nodes, replace=False)

        for node_id in node_ids:
            self._objective += self._local_search(node_id)

    def _local_search(self, node_index: int) -> float:
        """Performs a local search over the predecessors (out_node) of the given node.

        Parameters
        ----------
        node_index : int
            Node index.

        Returns
        -------
        float
            Change in objective function.
        """
        objective = 0.0
        current_out_node = self._predecessor_map[node_index]
        if current_out_node != -1:
            # partial removal of current node
            objective += (
                self._transition_weight(self._sub_out_map, current_out_node)
                - self._predecessor_weight[node_index]
            )
            self._selected_nodes[current_out_node] = False
            self._out_count[current_out_node] -= 1
            self._release_overlap(current_out_node)
        else:
            # add in node if it isn't in solution
            objective + self._transition_weight(self._add_in_map, node_index)

        # search local maximum
        max_obj_delta = np.finfo(np.float32).min
        argmax_obj_delta = -1
        argmax_weight = 0.0
        for i in range(
            self._in_out_digraph_indptr[node_index],
            self._in_out_digraph_indptr[node_index + 1],
        ):
            out_node = self._in_out_digraph_indices[i]
            if self._out_count[out_node] > 1 or self._forbidden[out_node]:
                continue

            weight = self._in_out_digraph_data[i]
            delta = self._transition_weight(self._add_out_map, out_node) + weight
            if delta > max_obj_delta:
                max_obj_delta = delta
                argmax_obj_delta = out_node
                argmax_weight = weight

        if objective + max_obj_delta <= 0.0:  # can't improve solution
            # return previous out node to solution
            if current_out_node != -1:
                self._forbid_overlap(current_out_node)
                self._selected_nodes[current_out_node] = True
                self._out_count[current_out_node] += 1

            return 0.0

        if current_out_node == -1:
            # add in node to solution
            self._in_count[node_index] += 1
            self._forbid_overlap(node_index)
            self._selected_nodes[node_index] = True

        # sanity check
        assert argmax_obj_delta != -1

        # adding to solution
        self._forbid_overlap(argmax_obj_delta)
        self._selected_nodes[argmax_obj_delta] = True
        self._out_count[argmax_obj_delta] += 1
        self._predecessor_map[node_index] = argmax_obj_delta
        self._predecessor_weight[node_index] = argmax_weight

        return objective + max_obj_delta

    def _forbid_overlap(self, node_index: int) -> None:
        """Mark nodes overlapping with the given index as forbidden."""
        row = self._overlap_indices[
            self._overlap_indptr[node_index] : self._overlap_indptr[node_index + 1]
        ]
        self._forbidden[row] = True

    def _release_overlap(self, node_index: int) -> None:
        """Releases forbidden mark of nodes overlappping with the given index."""
        row = self._overlap_indices[
            self._overlap_indptr[node_index] : self._overlap_indptr[node_index + 1]
        ]
        for i in row:
            indices = self._overlap_indices[
                self._overlap_indptr[i] : self._overlap_indptr[i + 1]
            ]
            self._forbidden[i] = np.any(self._selected_nodes[indices])

    def _transition_weight(self, transition: np.ndarray, node_id: int) -> float:
        """Computes the cost of a given transition for the given node.

        Parameters
        ----------
        transition : np.ndarray
            (2, 3, 3) transition matrix.
        node_id : int
            Node index.

        Returns
        -------
        float
            The cost (weight) of the transition.
        """
        weights = np.asarray(
            [
                self._appear_weight[node_id],
                self._disappear_weight[node_id],
                self._division_weight,
            ]
        )
        return np.sum(
            transition[self._in_count[node_id], self._out_count[node_id]] * weights
        )



import pickle
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
import zarr

from ultrack import segment
from ultrack.config.config import MainConfig
from ultrack.core.database import NodeDB, OverlapDB, get_node_values
from ultrack.core.segmentation import get_nodes_features


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        (
            {"segmentation.n_workers": 1, "segmentation.max_noise": 0.1},
            {"length": 4, "size": 128, "n_dim": 2},
        ),
        (
            {"segmentation.n_workers": 4, "segmentation.max_noise": 0.1},
            {"length": 4, "size": 128, "n_dim": 2},
        ),
        (
            {"segmentation.n_workers": 1, "segmentation.max_noise": 0.1},
            {"length": 4, "size": 64, "n_dim": 3},
        ),
        (
            {"segmentation.n_workers": 4, "segmentation.max_noise": 0.1},
            {"length": 4, "size": 64, "n_dim": 3},
        ),
    ],
    indirect=True,
)
def test_multiprocessing_segment(
    config_instance: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:
    foreground, contours, _ = timelapse_mock_data

    segment(
        foreground,
        contours,
        config_instance,
        properties=["centroid"],
    )

    assert config_instance.data_config.metadata["shape"] == list(contours.shape)

    df = pd.read_sql_table(
        NodeDB.__tablename__, con=config_instance.data_config.database_path
    )

    # assert all columns are present
    assert set(NodeDB.__table__.columns.keys()) == set(df.columns)

    nodes = {}
    # assert unpickling works
    for blob in df["pickle"]:
        node = pickle.loads(blob)
        nodes[node.id] = node

    overlaps = pd.read_sql_table(
        OverlapDB.__tablename__, con=config_instance.data_config.database_path
    )

    assert np.all(overlaps["node_id"] != overlaps["ancestor_id"])

    # asserting they really overlap
    for node_id, ancestor_id in zip(overlaps["node_id"], overlaps["ancestor_id"]):
        assert nodes[node_id].IoU(nodes[ancestor_id]) > 0.0

    overlaps = set(zip(overlaps["node_id"].tolist(), overlaps["ancestor_id"].tolist()))

    # assert we didn't miss any overlap
    for _, group in df.groupby(["t", "t_hier_id"]):
        indices = group["id"].tolist()
        for i in indices:
            node_i = nodes[i]
            for j in indices:
                if i == j or (i, j) in overlaps or (j, i) in overlaps:
                    # overlaps have alreayd been checked
                    continue
                node_j = nodes[j]
                assert node_i.IoU(node_j) == 0.0

    feats = get_node_values(config_instance.data_config, i, NodeDB.features)
    feats_name = config_instance.data_config.metadata["properties"]

    assert len(feats) == len(feats_name)
    assert isinstance(feats, np.ndarray)

    df = get_nodes_features(config_instance, include_persistence=True)

    assert (df["node_death"] >= df["node_birth"]).all()

    centroids_cols = [f"centroid-{i}" for i in range(contours.ndim - 1)]

    assert df.shape[0] == len(nodes)
    np.testing.assert_array_equal(
        sorted(df.columns.to_numpy(dtype=str)),
        sorted(
            ["t", "z", "y", "x", "area", "node_birth", "node_death"] + centroids_cols
        ),
    )



import pandas as pd
from rich import print
from rich.table import Table


def pretty_print_df(df: pd.DataFrame, title: str, row_name: str = "") -> None:
    """Converts dataframe to rich table and prints it."""
    table = Table(title=title)

    table.add_column(row_name)
    for c in df.columns:
        table.add_column(c)

    for name, row in df.iterrows():
        row = [str(value) for value in row]
        table.add_row(name, *row)

    print(table)



from typing import Optional

import click

from ultrack import solve
from ultrack.cli.utils import batch_index_option, config_option, overwrite_option
from ultrack.config import MainConfig


@click.command("solve")
@config_option()
@batch_index_option()
@overwrite_option()
def solve_cli(config: MainConfig, batch_index: Optional[int], overwrite: bool) -> None:
    """Compute tracks by selecting optimal linking between candidate segments."""
    solve(config, batch_index, overwrite)



from abc import ABC, abstractmethod
from typing import Literal, Optional

import pandas as pd
from numpy.typing import ArrayLike

from ultrack.config.config import TrackingConfig


class BaseSolver(ABC):
    def __init__(
        self,
        config: TrackingConfig,
    ) -> None:
        """
        Abstract base solver for api specification.

        Parameters
        ----------
        config : TrackingConfig
            Tracking configuration parameters.
        """
        self._config = config

    @abstractmethod
    def add_nodes(
        self,
        indices: ArrayLike,
        is_first_t: ArrayLike,
        is_last_t: ArrayLike,
        is_border: ArrayLike = False,
        node_prob: Optional[ArrayLike] = None,
    ) -> None:
        """Add nodes variables solver.

        Parameters
        ----------
        indices : ArrayLike
            Nodes indices.
        is_first_t : ArrayLike
            Boolean array indicating if it belongs to first time point and it won't receive appearance penalization.
        is_last_t : ArrayLike
            Boolean array indicating if it belongs to last time point and it won't receive disappearance penalization.
        is_border : ArrayLike
            Boolean array indicating if it belongs to the border and it won't receive (dis)apperance penalization.
            Default: False.
        node_prob: Optional[ArrayLike]
            If provided assigns a node probability score to the objective function.
        """

    @abstractmethod
    def add_edges(
        self, sources: ArrayLike, targets: ArrayLike, weights: ArrayLike
    ) -> None:
        """Add edges variables to solver.

        Parameters
        ----------
        source : ArrayLike
            Array of integers indicating source indices.
        targets : ArrayLike
            Array of integers indicating target indices.
        weights : ArrayLike
            Array of weights, input to the link function.
        """

    def set_standard_constraints(self) -> None:
        """Sets standard biological/flow constraints:
        - single incoming node (no fusion);
        - flow conservation (begin and end requires slack variables);
        - divisions only from existing nodes;
        """

    @abstractmethod
    def add_overlap_constraints(self, sources: ArrayLike, targets: ArrayLike) -> None:
        """Add constraints such that `source` and `target` can't be present in the same solution.

        Parameters
        ----------
        sources : ArrayLike
            Source nodes indices.
        targets : ArrayLike
            Target nodes indices.
        """

    @abstractmethod
    def enforce_nodes_solution_value(
        self,
        indices: ArrayLike,
        variable: Literal["appear", "disappear", "division", "node"],
        value: bool,
    ) -> None:
        """Constraints given nodes' variables to 1.

        Parameters
        ----------
        indices : ArrayLike
            Nodes indices.
        variable : str
            Slack variable to constraint.
        value : bool
            Value to constraint to.
        """

    @abstractmethod
    def optimize(self) -> float:
        """
        Computes solution.

        Returns
        -------
        float
            Solution objective value.
        """

    @abstractmethod
    def solution(self) -> pd.DataFrame:
        """Returns the nodes present on the solution.

        Returns
        -------
        pd.DataFrame
            Dataframe indexed by nodes as indices and their parent (NA if orphan).
        """



from pathlib import Path
from typing import Optional, Tuple

import click
from napari.plugins import _initialize_plugins
from napari.viewer import ViewerModel
from tifffile import imread
from tqdm import tqdm

from ultrack.cli.utils import (
    config_option,
    napari_reader_option,
    output_directory_option,
    overwrite_option,
    tuple_callback,
)
from ultrack.config import MainConfig
from ultrack.core.export import to_ctc, to_trackmate, to_tracks_layer, tracks_to_zarr
from ultrack.core.solve.sqltracking import SQLTracking
from ultrack.imgproc.measure import tracks_properties
from ultrack.utils.data import validate_and_overwrite_path


@click.command("ctc")
@output_directory_option(
    help="Output directory to save segmentation and lineage graph (e.g. 01_RES)."
)
@config_option()
@overwrite_option()
@click.option(
    "--margin",
    "-ma",
    default=0,
    type=int,
    show_default=True,
    help="Ignored margin on xy-plane.",
)
@click.option(
    "--scale",
    "-s",
    default=None,
    type=str,
    show_default=True,
    help="Output scale factor (e.g. 0.2,1,1 ). Useful when tracking was done on upscaled input."
    "Must have length 3, first dimension is ignored on for 2-d images.",
    callback=tuple_callback(length=3, dtype=float),
)
@click.option(
    "--first-frame-path",
    default=None,
    type=click.Path(path_type=Path, exists=True),
    show_default=True,
    help="Optional first frame path used to select a subset of lineages connected to this reference annotations.",
)
@click.option(
    "--dilation-iters",
    "-di",
    default=0,
    type=int,
    show_default=True,
    help="Iterations of radius 1 morphological dilations on labels, applied after scaling.",
)
@click.option(
    "--stitch-tracks",
    default=False,
    is_flag=True,
    type=bool,
    help="Stitches (connects) incomplete tracks nearby tracks on subsequent time point.",
)
def ctc_cli(
    output_directory: Path,
    config: MainConfig,
    margin: int,
    scale: Optional[Tuple[float]],
    first_frame_path: Optional[Path],
    dilation_iters: int,
    stitch_tracks: bool,
    overwrite: bool,
) -> None:
    """Exports tracking results to cell-tracking challenge (http://celltrackingchallenge.net) format."""

    if first_frame_path is None:
        first_frame = None
    else:
        first_frame = imread(first_frame_path)

    to_ctc(
        output_directory,
        config,
        margin=margin,
        scale=scale,
        first_frame=first_frame,
        dilation_iters=dilation_iters,
        stitch_tracks=stitch_tracks,
        overwrite=overwrite,
    )


@click.command("zarr-napari")
@output_directory_option(
    help="Output directory to save segmentation masks and tracks table (e.g. results)."
)
@config_option()
@overwrite_option()
@click.option(
    "--measure",
    default=False,
    show_default=True,
    is_flag=True,
    help="Add segmentation measurements to tracks table.",
)
@click.option(
    "--image-path",
    "-i",
    type=click.Path(path_type=Path, exists=True),
    default=None,
    help="If provided, tracks measurements will be enriched with intensity-based properties.",
)
@napari_reader_option()
def zarr_napari_cli(
    output_directory: Path,
    config: MainConfig,
    overwrite: bool,
    measure: bool,
    image_path: Optional[Path],
    reader_plugin: str,
) -> None:
    """
    Exports segments to zarr and tracks to napari tabular format
    (.csv for tracklets, parent relationship is lost).
    """
    tracks_path = output_directory / "tracks.csv"
    validate_and_overwrite_path(tracks_path, overwrite, "cli")

    segm_path = output_directory / "segments.zarr"
    validate_and_overwrite_path(segm_path, overwrite, "cli")

    output_directory.mkdir(exist_ok=True)

    tracks, _ = to_tracks_layer(config, include_parents=True)
    tracks.to_csv(tracks_path, index=False)  # saving before measures just to be sure

    segments = tracks_to_zarr(config, tracks, store_or_path=segm_path)

    if measure or image_path is not None:
        # extract segmentation measurements
        if image_path is None:
            image = None
        else:
            _initialize_plugins()
            viewer = ViewerModel()
            image = [
                layer.data[0] if layer.multiscale else layer.data
                for layer in viewer.open(image_path, plugin=reader_plugin)
            ]
            del viewer

        tracks_w_measures = tracks_properties(
            segments=segments,
            tracks_df=tracks,
            image=image,
            scale=config.data_config.metadata.get("scale"),
            n_workers=config.data_config.n_workers,
        )
        tracks_w_measures.to_csv(tracks_path, index=False)


@click.command("trackmate")
@click.option(
    "--output-path",
    "-o",
    required=True,
    type=click.Path(path_type=Path),
    show_default=True,
    help="TrackMate XML output path.",
)
@config_option()
@overwrite_option()
def trackmate_cli(
    config: MainConfig,
    output_path: Path,
    overwrite: bool,
) -> None:
    """
    Exports tracking results to TrackMate XML format.
    """
    to_trackmate(config, output_path, overwrite)


@click.command("lp")
@click.option(
    "--output-path",
    "-o",
    required=True,
    type=click.Path(path_type=Path),
    show_default=True,
    help="ILP model (.lp) output path.",
)
@config_option()
@overwrite_option()
def lp_cli(
    config: MainConfig,
    output_path: Path,
    overwrite: bool,
) -> None:
    """
    Exports tracking ILP to .lp format.
    """
    tracker = SQLTracking(config)

    for batch_index in tqdm(range(tracker.num_batches), "Exporting ILP"):
        tracker.construct_model(index=batch_index)

        if tracker.num_batches == 1:
            model_path = output_path
        else:
            name = f"{output_path.name.removesuffix('.lp')}_{batch_index:02d}.lp"
            model_path = output_path.parent / name

        if model_path.exists() and not overwrite:
            raise ValueError(f"File {model_path} already exists. Set `--overwrite`.")

        tracker.solver._model.write(str(model_path))


@click.group("export")
def export_cli() -> None:
    """Exports tracking and segmentation results to selected format."""


export_cli.add_command(ctc_cli)
export_cli.add_command(lp_cli)
export_cli.add_command(trackmate_cli)
export_cli.add_command(zarr_napari_cli)



import logging
from types import ModuleType
from typing import Optional

import numpy as np
import pytest
import scipy.ndimage as ndi
from skimage.data import cells3d
from skimage.morphology import reconstruction

from ultrack.imgproc import Cellpose, detect_foreground, inverted_edt, robust_invert
from ultrack.imgproc.segmentation import reconstruction_by_dilation
from ultrack.utils.cuda import to_cpu, xp

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize("np_module,channel_axis", [(xp, None), (xp, 0), (np, 2)])
def test_foreground_detection(
    np_module: ModuleType,
    channel_axis: Optional[int],
    request,
) -> None:

    cells = cells3d()
    nuclei = np_module.asarray(cells[:, 1])
    membrane = np_module.asarray(cells[:, 0])

    if channel_axis is not None:
        nuclei = np_module.stack([nuclei] * 2, axis=channel_axis)
        membrane = np_module.stack([membrane] * 2, axis=channel_axis)

    contours = robust_invert(membrane, [1, 1, 1], channel_axis=channel_axis)
    foreground = detect_foreground(
        nuclei, [1, 1, 1], sigma=50, channel_axis=channel_axis
    )

    assert contours.min() == 0.0
    assert contours.max() == 1.0

    if request.config.getoption("--show-napari-viewer"):
        import napari

        viewer = napari.Viewer()

        viewer.add_image(cells, blending="additive", channel_axis=1)
        viewer.add_labels(to_cpu(foreground))
        viewer.add_image(
            to_cpu(contours), blending="additive", colormap="gray_r", rendering="minip"
        )

        napari.run()


def test_reconstruction_by_dilation(request) -> None:
    cells = cells3d()
    membrane = cells[cells.shape[0] // 2, 0]

    seed = ndi.gaussian_filter(membrane, 5)

    iterative_bkg = reconstruction_by_dilation(seed, membrane, 250)
    exact_bkg = reconstruction(seed, membrane, method="dilation")

    if request.config.getoption("--show-napari-viewer"):
        import napari

        viewer = napari.Viewer()

        viewer.add_image(membrane, blending="additive", visible=False)
        viewer.add_image(seed, blending="additive", visible=False)
        viewer.add_image(iterative_bkg, blending="additive", colormap="red")
        viewer.add_image(exact_bkg, blending="additive", colormap="green")

        napari.run()

    np.testing.assert_array_almost_equal(iterative_bkg, exact_bkg)


def test_inverted_edt() -> None:
    mask = np.array([[0, 0, 1], [0, 1, 1], [1, 1, 1]], dtype=bool)
    expected_output = np.array(
        [[1.0, 1.0, 0.292893], [1.0, 0.646447, 0.209431], [0.646447, 0.292893, 0.0]]
    )
    output = inverted_edt(mask, voxel_size=(1, 2))
    np.testing.assert_array_almost_equal(output, expected_output)


def test_cellpose() -> None:
    pytest.importorskip("cellpose")
    image = np.random.rand(100, 100)
    cellpose_model = Cellpose()
    cellpose_model(image)
